import configparser
import json
import regex
import os
import importlib_resources

class SOAPConfig():

    def __init__(self, config_file):
        cfg = configparser.ConfigParser()
        my_resources = importlib_resources.files("soapcw")
        cfg.read((my_resources / "cnn"/ "init_files" / "default_gauss.ini"))

        cfg.read(config_file)

        self.config_file = os.path.abspath(config_file)
        self.float_list = ["band_starts","band_ends","band_widths"]
        self.int_list = ["strides","fc_layers","img_dim", "avg_pool_size"]
        self.string_list = ["load_directory","save_options","type","band_types","sft_dirs"]
        self.tuple_list = ["conv_layers"]
        self.floats = ["band_load_size", "snr_width_line", "snr_width_signal", 
                        "prob_line", "left_right_prob", "det1_prob", "det2_prob",
                        "snrmin","snrmax","learning_rate","data_load_size","tstart","dropout"]
        self.ints = ["memory", "request_disk", "n_jobs", "n_channels","nperband",
                     "n_summed_sfts","n_epochs","save_interval","nsfts", "tstart", "tend",
                     "n_train_multi_size","fdim","input_dim","par_dim","num_predict_params","latent_dim","stride"]
        self.bools = ["resize_image", "overwrite_files", "gen_noise_only"]
        self.strings = ["save_dir", "narrowband_sft_dir", "accounting_group", "root_dir", 
                        "run", "type", "lookup_dir", "model_type","search_exec","dist_type"]


        self.config = self.parse_config(cfg)

    def __getitem__(self, key):
        return self.config[key]

    def load_list(self, val, partype):
        if "," in val:
            val = val.strip("[").strip("]").strip("(").strip(")").split(",")
        else:
            val = [val.strip("[").strip("]").strip("(").strip(")").strip(",")]
        
        if partype == "float":
            val = [float(v) for v in val]
        elif partype == "int":
            val = [int(v) for v in val]
        elif partype == "string":
            val = [v.replace('"','').replace(' ','') for v in val]
        else:
            raise Exception(f"Type {partype} not supported")

        return val

    def get_bool(self, val):
        if val in ["false", "False", False, 0]:
            return False
        elif val in ["true", "True", True, 1]:
            return True
        else:
            raise Exception("Value not of Bool type")
            
    def parse_config(self, cfg):

        parsed_dict = {}
        for key, val in cfg.items():
            parsed_dict[key] = {}
            for key2, val2 in val.items():
                # if comma then part of a list
                if val2 in ["none", "None", None]:
                    parsed_dict[key][key2] = None
                elif key2 in self.float_list:
                    parsed_dict[key][key2] = self.load_list(val2, "float")
                elif key2 in self.int_list:
                    parsed_dict[key][key2] = self.load_list(val2, "int")
                elif key2 in self.string_list:
                    parsed_dict[key][key2] = self.load_list(val2, "string")
                elif key2 in self.floats:
                    parsed_dict[key][key2] = float(val2)
                elif key2 in self.ints:
                    parsed_dict[key][key2] = int(val2)
                elif key2 in self.bools:
                    parsed_dict[key][key2] = self.get_bool(val2)
                elif key2 in self.tuple_list:
                    temp_out = []
                    for mod in regex.split(r"\s*,\s*(?![^(]*\))", val2.strip("[").strip("]")):
                        if mod == "":
                            continue
                        out_tuple = tuple([int(vl) for vl in mod.strip("\n").strip("(").strip(")").split(",")])
    
                        temp_out.append(out_tuple)

                    parsed_dict[key][key2] = temp_out
                elif key2 in self.strings:
                    parsed_dict[key][key2] = val2.replace('"','')
                else:
                    print(f"Key not in defaults: {key}, {key2}")
                    if val2 in ["none", "None"]:
                        parsed_dict[key][key2] = None
                    else:
                        parsed_dict[key][key2] = val2.replace('"','')
                
        return parsed_dict



#!/home/joseph.bayley/.virtualenvs/soap36/bin/python
import matplotlib
matplotlib.use("agg")
import numpy as np
import sys
import os
#import plot
import pickle
import time
import datetime
import soapcw as soap
from soapcw import cw
import argparse
from skimage.transform import resize
import skimage.measure
from scipy.signal import butter,filtfilt
import matplotlib.pyplot as plt
import copy
from soapcw.cnn import generate_train_data
from soapcw.cnn import generate_test_data
#import mdc_test
from soapcw.cnn import generate_gaussian_train_data


def find_sft_file(indir,bandmin,bandmax):
    """
    find the narrowbanded sft file in a directory which runs between band min and band max
    args
    --------------
    indir: string
        search directory
    bandmin: float
        minimum frequency
    bandmax: float
        maximum frequency

    returns
    ----------------
    sft file that matches this string
    sft file patter : H-<nsft>_<det>_<tsft>SFT_NB_F<freq>Hz0_W<width>Hz180-<sttime>-<span>.sft
    """
    # lists all files in directory
    hindir = os.path.join(indir,"H1")
    lindir = os.path.join(indir,"L1")

    hname = None
    lname = None

    # loops over files to find one with matching string
    for fname in os.listdir(hindir):
        #print("F{:0>4}Hz".format(int(bandmin)), fname)
        if "F{:0>4}Hz".format(int(bandmin)) in fname:
            hname = fname

    for fname in os.listdir(lindir):
        #if "F{}_{}.sft".format(bandmin,bandmax) in fname:
        if "F{:0>4}Hz".format(int(bandmin)) in fname:
            lname = fname

    if hname is None or lname is None:
        print("NO SFT FOUND {}, {}, {}, F{:0>4}Hz".format(indir, bandmin, bandmax, int(bandmin)))

    return os.path.join(hindir,hname),os.path.join(lindir,lname)


def return_outputs(datah,datal,fmin,fmax,snr,out_snr="./",h0=None,depth=None,size=(156,89),resize_image=True,test_plot=False,pars=None,test=False,epochs=None,minmaxpaths=None,inj_track=None, save_options = ["vit_imgs","H_imgs","L_imgs","stats","sum_stats","pars","diffs","paths","powers","plots"], lookup_dir = "/home/joseph.bayley/data/soap/lookup_tables/line_aware_optimised/", degfree = 96, brange = "", totshape = None):
    """
    save the outputs of the SOAP search and the spectrograms of the data
    
    """
    #save_options = ["stats","paths", "pars", "powers"]
    #save_options = ["stats","paths", "pars", "powers","vit_imgs","H_imgs","L_imgs"]
    # define all save directories for images, stats, pars, plots etc. out_snr only exists for training data
    """
    outdirs = {}
    for dirname in save_options:
        outdirs[dirname] = os.path.join(*[outpath,dirname,brange,out_snr])
    """
    #print(outdirs)

    # set location of lookup table directory
    #lookup_dir = "/home/joseph.bayley/data/soap/lookup_tables/lookup_both_opt_gauss_2/"
    
    # for line veto, choose optimum parameters from previous searches
    #lookup_1 = str(lookup_dir) + "/log_signoiseline_1det_{}degfree_4.0_5.0_0.0387.pkl".format(degfree,4.0,5.0,0.0387)
    #lookup_2 = str(lookup_dir) + "/log_signoiseline_2det_{}degfree_4.0_5.0_0.0387.pkl".format(degfree,4.0,5.0,0.0387)

    #print("SOAP pars:",fmin,degfree, np.mean(datah), np.mean(datal))
    lookup_1 = str(lookup_dir) + "/log_signoiseline_1det_{}degfree_{}_{}_{}.pkl".format(degfree,4.0,10.0,0.4)
    lookup_2 = str(lookup_dir) + "/log_signoiseline_2det_{}degfree_{}_{}_{}.pkl".format(degfree,4.0,10.0,0.4)

    # Gaussian lookup parmaeters
    #lookup_1 = str(lookup_dir) + "/signoiseline_1det_{}_{}_{}.txt".format(2.1,1.0,0.0)
    #lookup_2 = str(lookup_dir) + "/signoiseline_2det_{}_{}_{}.txt".format(2.1,1.0,0.0)

    # set the transition matrix as likely to be in the same bin in each detector and slightyl more probable to go straight (mostly so it goest straight through gaps)
    #tr = soap.tools.transition_matrix_2d_5(1.0000001,1.000001,1e400,1e400,log=True)
    tr = soap.tools.transition_matrix_2d(1.00000001,1e400,1e400)


    if "stats" in save_options:
        # run two detector search using the line aware statistic
        viterbi = soap.two_detector(tr,datah,datal,lookup_table_2det=lookup_2,lookup_table_1det=lookup_1)

    roll_track_sub = viterbi.vit_track - np.roll(viterbi.vit_track, 1)
    if np.any(np.abs(roll_track_sub[3:-3]) > 3):
        print("diff of track >1")
        print(viterbi.vit_track)
        sys.exit()
        

    # summed stat
    if "sum_stats" in save_options:
        sum_viterbi = soap.two_detector(tr,datah,datal)
    
    if inj_track is not None and "diffs" in save_options:
        diffs = np.array(viterbi.vit_track1) - np.array(inj_track)[:len(viterbi.vit_track1)]

    # get the normalised viterbi feather plots
    vit_data = viterbi.vitmap

    # get the normalised SFT power along the optimum track
    if "powers" in save_options:
        powers = soap.tools.track_power(viterbi.vit_track1,datah),soap.tools.track_power(viterbi.vit_track2,datal)

    #create the figure showing SFTs and tracks
    if "plots" in save_options:
        plotfig = make_figure(fmin,fmax,epochs[0],epochs[-1],datah,datal,[viterbi.vit_track1,viterbi.vit_track2],powers,vit_data,1800.,plot_track=True)

    # create the appropriate save directories
    """
    for key,j in outdirs.items():
        if not os.path.isdir(j):
            os.makedirs(j)
    """
    # if a size is not set then set the size of output image to 1/3 length in time and 1/2 in frequency
    if size is None:
        size = (vit_data.shape[0] / 3, vit_data.shape[1] / 2)
        
    # resize the normalised SFTs and the viterbi maps using skimages resize function which interpolates (also option for maxpooling)
    if resize_image:
            
        # viterbi map resize
        vit_resize      = resize(vit_data, size, anti_aliasing=True)
        #vit_resize      = skimage.measure.block_reduce(vit_data, (3,2), np.max)
        vit_resize_norm = vit_resize#/vit_resize.max()

        # H1 sfts resize
        deth_resize      = resize(datah, size, anti_aliasing=True)
        #deth_resize      = skimage.measure.block_reduce(datah, (3,2), np.max)
        deth_resize_norm = deth_resize#/deth_resize.max()

        # L1 sfts resize
        detl_resize      = resize(datal, size, anti_aliasing=True)
        #detl_resize      = skimage.measure.block_reduce(datal, (3,2), np.max)
        detl_resize_norm = detl_resize#/detl_resize.max()
    
    else:
        # in this case dont resize the image atall
        vit_resize_norm  = vit_data
        deth_resize_norm = datah
        detl_resize_norm = datal
    
    # if images from same data set are not the same size the raise error and exit
    if np.shape(vit_resize_norm) != np.shape(deth_resize_norm) or np.shape(vit_resize_norm) != np.shape(detl_resize_norm):
        print("Wrong Sizes")
        sys.exit()

    if totshape is None:
        totshape = (np.shape(vit_resize_norm), np.shape(vit_data))
    else:
        if totshape[1] != np.shape(vit_data):
            print("TOTALSHAPE DIFFERENT",totshape[1],np.shape(vit_data))
        if totshape[0] != np.shape(vit_resize_norm):
            print("RESIZESHAPE DIFFERENT",totshape[0],np.shape(vit_resize_norm))
            raise Exception("Shapes different between runs")


    return_dict = {}
    if "vit_imgs" in save_options:
        return_dict["vit_imgs"] = vit_resize_norm
        
    if "stats" in save_options:
        return_dict["stats"] = viterbi.max_end_prob
            
    if "sum_stats" in save_options:
        return_dict["sum_stats"] = sum_viterbi.max_end_prob
            
    if "H_imgs" in save_options:
        return_dict["H_imgs"] = deth_resize_norm
    if "L_imgs" in save_options:
        return_dict["L_imgs"] = detl_resize_norm

    if "paths" in save_options:
        return_dict["paths"] = [epochs,viterbi.vit_track]
    if "powers" in save_options:
        return_dict["powers"] = [epochs,powers[0],powers[1]]

    # save parametrs of injections if there are injections
    if pars is not None:
        return_dict["pars"] = pars

    return return_dict

def make_figure(flow,fhigh,tstart,tend,sum_obs1,sum_obs2,tracks,powers,vit_m,tsft,plot_track=True):
    fig, ax = plt.subplots(nrows=4,figsize = (14,16))
    im1 = ax[0].imshow(sum_obs1.T,aspect='auto',origin='lower',extent=[tstart,tend,flow,fhigh],label = "H1",cmap = plt.get_cmap("Greys"))
    im2 = ax[1].imshow(sum_obs2.T,aspect='auto',origin='lower',extent=[tstart,tend,flow,fhigh],label='L1',cmap = plt.get_cmap("Greys"))
    vmin,vmax = 0,1
    vim = ax[3].imshow(vit_m.T,aspect='auto',origin='lower',extent=[tstart,tend,flow,fhigh],cmap = plt.get_cmap('Greys'),label = "V")

    if plot_track:
        ax[0].plot(np.linspace(tstart,tend,len(tracks[0])),tracks[0]/tsft + flow,'C0',label = "Viterbi track")
        ax[1].plot(np.linspace(tstart,tend,len(tracks[1])),tracks[1]/tsft + flow,'C2',label = "Viterbi track")
    else:
        pass

    ax[2].plot(np.linspace(tstart,tend,len(powers[0])),powers[0],'C0',label = "H1")
    ax[2].plot(np.linspace(tstart,tend,len(powers[1])),powers[1],'C2',label = "L1")

    ax[0].set_xticklabels([])
    ax[1].set_xticklabels([])
    ax[3].set_xlabel("Time [s]",fontsize=20)
    leg0 = ax[0].legend(fancybox=True,fontsize=19,loc = 'upper right',borderpad = 0.2)
    leg1 = ax[1].legend(fancybox=True,fontsize=19,loc = 'upper right',borderpad = 0.2)
    leg2 = ax[2].legend(fancybox=True,fontsize=19,loc = 'upper right',borderpad = 0.2)
    leg3 = ax[3].legend(fancybox=True,fontsize=19,loc = 'upper right',borderpad = 0.2)

    ax[0].text(0.02, 0.85, 'H1',
               verticalalignment='bottom', horizontalalignment='left',
               transform=ax[0].transAxes,
               color='C0', fontsize=20,bbox={'facecolor':'white', 'alpha':0.5, 'pad':3,'boxstyle':'round,pad=0.3','ec':'none'})
    ax[1].text(0.02, 0.85, 'L1',
               verticalalignment='bottom', horizontalalignment='left',
               transform=ax[1].transAxes,
               color='C2', fontsize=20,bbox={'facecolor':'white', 'alpha':0.5, 'pad':3,'boxstyle':'round,pad=0.3','ec':'none'})
    ax[2].text(0.02, 0.85, 'Viterbi',
               verticalalignment='bottom', horizontalalignment='left',
               transform=ax[2].transAxes,
               color='black', fontsize=20,bbox={'facecolor':'white', 'alpha':0.5, 'pad':3,'boxstyle':'round,pad=0.3','ec':'none'})
    #divider = make_axes_locatable(ax[2])                                                                                                                                                      
    #cax = divider.append_axes("right", size="2%", pad=0.05,)                                                                                                                                  
    cax = fig.add_axes([0.91, 0.11, 0.01, 0.19])
    cbar = plt.colorbar(vim, cax=cax)
    cax1 = fig.add_axes([0.91, 0.69, 0.01, 0.19])
    cbar1 = plt.colorbar(im1, cax=cax1)
    cax2 = fig.add_axes([0.91, 0.496, 0.01, 0.19])
    cbar2 = plt.colorbar(im2, cax=cax2)

    cbar.ax.tick_params(labelsize=16)
    cbar1.ax.tick_params(labelsize=16)
    cbar2.ax.tick_params(labelsize=16)
    cbar.set_label(label=r"$\propto \log\left[O_{\rm{SGL}}\right]$", fontsize = 20)
    cbar1.set_label(label=r"SFT power", fontsize = 20)
    cbar2.set_label(label=r"SFT power", fontsize = 20)
    ax[0].tick_params(axis='both', which='major', labelsize=17)
    ax[1].tick_params(axis='both', which='major', labelsize=17)
    ax[2].tick_params(axis='both', which='major', labelsize=17)
    ax[3].tick_params(axis='both', which='major', labelsize=17)
    ax[0].yaxis.set_ticks(np.arange(flow+0.01, fhigh-0.01, 0.02))
    ax[1].yaxis.set_ticks(np.arange(flow+0.01, fhigh-0.01, 0.02))
    #ax[2].yaxis.set_ticks(np.arange(flow+0.01, fhigh-0.01, 0.02))                                                                                                                             
    ax[3].yaxis.set_ticks(np.arange(flow+0.01, fhigh-0.01, 0.02))
    ax[0].set_ylim([flow,fhigh])
    ax[1].set_ylim([flow,fhigh])
    #ax[2].set_ylim([flow,fhigh])                                                                                                                                                              
    ax[3].set_ylim([flow,fhigh])
    ax[0].set_xlim([tstart,tend])
    ax[1].set_xlim([tstart,tend])
    ax[2].set_xlim([tstart,tend])
    ax[3].set_xlim([tstart,tend])
    ax[0].set_ylabel("Frequency [Hz]",fontsize=20)
    ax[1].set_ylabel("Frequency [Hz]",fontsize=20)
    ax[2].set_ylabel("Normalised \n SFT Power",fontsize=20)
    ax[3].set_ylabel("Frequency [Hz]",fontsize=20)
    ax[0].get_yaxis().get_major_formatter().set_useOffset(False)
    ax[1].get_yaxis().get_major_formatter().set_useOffset(False)
    ax[3].get_yaxis().get_major_formatter().set_useOffset(False)
    fig.subplots_adjust(wspace=0, hspace=0)
    return fig


            
def sum_sft(data,nsfts=48):

    data_av1 = []
    fraction = []
    for i in np.linspace(0,len(data)-nsfts-1,len(data)-nsfts)[::nsfts]:
        end = i+nsfts
        chunk = data[int(i):int(end)]
        av = np.nansum(chunk.T,axis=1)
        data_av1.append(av)
                        
    return np.array(data_av1)


def str2bool(v):
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')



def main():
    if len(sys.argv) < 2 or sys.argv[1] == '-h':
        print(sys.argv)
        sys.exit(0)
    else:
        parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser.add_argument('-v', '--verbose', help='display status', action='store_true')
        parser.add_argument("-c", "--config-file", help="config file contatining parameters")
        parser.add_argument('-l', '--band-min', help='lower frequency of band', type=float, required=True)
        parser.add_argument('-u', '--band-max', help='upper frequency of band', type=float, required=True)
        parser.add_argument('-w', '--band-width', help='width of indidual band', type=float, required=True)
        parser.add_argument('-dt', '--data-type', help='train, validation or test data', type=str, default = "train")
        parser.add_argument('-rt', '--run-type', help='run type, gauss: return gaussian noise, else:O1,O2,O3,(O4)',required=True)
        parser.add_argument('-sr', '--search', help='generate search data, if true generates data with no injections', type=str2bool, const=True, default = False,nargs='?')
        parser.add_argument('-nn', '--no-noise', help='do not regenerate noise bands', type=str2bool, const=True, default = False,nargs='?')
        parser.add_argument('-so', '--save-options', help='files to save', type=str, default = "vit_img stats pars H_imgs L_imgs vit_imgs",nargs="?")
        parser.add_argument('-np', '--nperband', help='numnber of examples per band gaussian noise', type=int, default = 20,nargs="?")

        try:                                                     
            args = parser.parse_args()  
        except:  
            sys.exit(1)

        from soapcw.soap_config_parser import SOAPConfig

        if args.config_file is not None:
            config = SOAPConfig(args.config_file)
        else:
            config = None


        if args.data_type == "test":
            save_dir = os.path.join(config["general"]["save_dir"], "test")
            if args.run_type != "gaussian":
                generate_test_data.loop_band_test_data(args.load_dir,save_dir,args.band_min,args.band_max,args.band_width,args.resize_image, num_repeat = args.nperband)
            elif args.run_type == "gaussian":
                gaussian_train.loop_band_train_augment(loadpath=args.load_dir,path=save_dir,bandmin=args.band_min,bandmax=args.band_max, band_width=args.band_width,resize_image=args.resize_image, nonoise=args.no_noise,snr_min=args.snr_min,snr_max=args.snr_max,nperband=1,test=True,save_options=args.save_options.split(" "))

        elif args.data_type in ["train", "validation"]:
            save_dir = os.path.join(config["general"]["save_dir"], args.data_type)

            print("runtype", args.run_type)
            if args.run_type == "gaussian" or args.run_type == "gauss":
                generate_gaussian_train_data.loop_band_train_augment(
                    config, 
                    path=save_dir,
                    bandmin=args.band_min,
                    bandmax=args.band_max, 
                    band_width=args.band_width, 
                    nonoise=args.no_noise,
                    snr_min=config["data"]["snrmin"],
                    snr_max=config["data"]["snrmax"],
                    nperband=args.nperband,
                    test=False,
                    save_options=config["data"]["save_options"]
                    )
            else:
                generate_train_data.loop_band_train_augment(
                    config, 
                    save_dir,
                    args.band_min,
                    args.band_max,
                    args.band_width,
                    config["data"]["resize_image"], 
                    gen_noise_only=config["data"]["gen_noise_only"],
                    snr_min=config["data"]["snrmin"],
                    snr_max=config["data"]["snrmax"],
                    save_options=config["data"]["save_options"]
                    )

        """
        elif args.mdc == True:
            mdc_test.loop_band_mdc(args.save_dir,args.load_dir,args.noise_dir,args.band_min,args.band_max,args.band_width,args.resize_image)
        elif args.search == True:
            generate_test_data.loop_band_test_data(args.load_dir,args.save_dir,args.band_min,args.band_max,args.band_width,args.resize_image,search = True)
        """

if __name__ == '__main__':       
    main()




from soapcw.cnn import cnn_data_gen
import numpy as np
from soapcw import cw
import time
import os
import datetime
import pickle 
import sys
import h5py

def loop_band_train_augment(config, path,bandmin,bandmax,band_width,loadpath=None,nonoise=False,snr_min = 50,snr_max=150,nperband=10,test=False, save_options=None):
    """
    loop over bands to create training data
    """

    tsft = 1800.
    for i, (minband, maxband) in enumerate(zip(config["data"]["band_starts"], config["data"]["band_ends"])):
        if minband <= bandmin < maxband:
            stride = config["data"]["strides"][i]
            degfree = int(2 * config["data"]["n_summed_sfts"] * stride)
            width = config["data"]["band_widths"][i]
            nbin = 180#int(width*tsft)
            plotcut = stride * nbin
            brange = f"band_{int(minband):d}_{int(maxband):d}"
            break
        else:
            continue


    start_time = time.time()
    # in 2Hz band, set frequencies and indicies with 0.1 Hz bands not overlapping
    range_bands = np.round(np.arange(bandmin,bandmax-width,width),1)
    range_index = np.arange(0,(bandmax-width-bandmin)*int(tsft),180*stride).astype(int)

    # set which bands are odd and even with and ondex of 1 or 0
    odd_even = np.zeros(len(range_bands))
    odd_even[1::2] = 1
    
    tmin, tmax = None,None#931035615.,931035615.+1800*100

    # set integer Hz as vetos, this will cut out a few bins surrounding
    vetolist = list(np.arange(bandmin,bandmax))

    # find the appropriate sft files
    if config["general"]["load_dir"] is not None:
        hname,lname = cnn_data_gen.find_sft_file(config["general"]["load_dir"],bandmin,bandmax)
        
        # load in the narrowbanded sfts, normalise them, fill in the gaps and save the running median
        sfts_noise = cw.LoadSFT("{};{}".format(hname,lname),norm=True,filled=True,vetolist = vetolist,save_rngmed=True)
        
        noise_est = sfts_noise.H1.get_sh(pos="median"),sfts_noise.L1.get_sh(pos="median")

        # sum the sfts
        sfts_noise.sum_sfts()

        nsft = len(noise_est[0])
        tstart = sfts_noise.H1.epochs[0]

    else:
        noise_est = 1,1 
        nsft = config["data"]["nsfts"]
        tstart = config["data"]["tstart"]

    # make appropriate directoes
    if not os.path.isdir(path):
        os.makedirs(path)
    erange = 15

    snrrange = [snr_min,snr_max]
    
    noise_out_snr = "snr_0.0_0.0"
    signal_out_snr = "snr_{}_{}".format(snrrange[0],snrrange[1])
    #save_options = ["stats","paths", "pars", "powers"]
    #save_options = ["stats","paths", "pars", "powers","vit_imgs","H_imgs","L_imgs"]
    # define all save directories for images, stats, pars, plots etc. out_snr only exists for training data

    even_noise_filenames = {}
    even_signal_filenames = {}
    odd_noise_filenames = {}
    odd_signal_filenames = {}

    even_noise_save_outs = {}
    even_signal_save_outs = {}
    odd_noise_save_outs = {}
    odd_signal_save_outs = {}

    for dirname in save_options:

        even_noise_save_outs[dirname] = []
        even_signal_save_outs[dirname] = []
        odd_noise_save_outs[dirname] = []
        odd_signal_save_outs[dirname] = []

                                                     
    
    # main loop over all sub bands
    for i in range(len(range_bands)):
        
        # set initial band frequency
        k = range_bands[i]

        ms = np.round((range_bands[i] - bandmin)*1800).astype(int)
        #snr_start,snr_end = 80,150
        #ms = int(range_index[i])
        me = int(ms+180*stride)
        
        # put odd bands into one folder and even bands into another folder
        if odd_even[i] == 1:
            split = "odd"
        else:
            split = "even"

        outpath = os.path.join(path,split)                     
        print("inds",ms,me)
        # get the median of the running medians as an estimate of the noise floow in this band.
        if loadpath is not None:
            av_sh = [np.nanmedian(sfts_noise.H1.rng_med[:,ms:me],axis=1),np.nanmedian(sfts_noise.L1.rng_med[:,ms:me],axis=1)]
        else:
            av_sh = [1,1]
                
        # set lower frequency
        #flow = np.round(k + ms / 1800. , 2)
        flow = k

        # run injections function with new data, this saves noise iamge and injects signal and saves signal images
        for i in range(nperband):
            noise_outputs, signal_outputs = run_and_inj(flow,width,outpath,resize_image=config["data"]["resize_image"],av_sh=av_sh,snrrange=snrrange,test=test, save_options=save_options, nsft = nsft, tstart = tstart, stride = stride, degfree = degfree, brange = brange)
            for key in save_options:
                if split == "even":
                    even_noise_save_outs[key].append(noise_outputs[key])
                    even_signal_save_outs[key].append(signal_outputs[key])
                elif split == "odd":
                    odd_noise_save_outs[key].append(noise_outputs[key])
                    odd_signal_save_outs[key].append(signal_outputs[key])

             
    even_noise_filenames = os.path.join(*[path,"even",brange,noise_out_snr,f"freq_{bandmin}_{bandmax}_{nperband*len(range_bands)}.hdf5"])
    even_signal_filenames = os.path.join(*[path,"even",brange,signal_out_snr,f"freq_{bandmin}_{bandmax}_{nperband*len(range_bands)}.hdf5"])
    odd_noise_filenames = os.path.join(*[path,"odd",brange,noise_out_snr,f"freq_{bandmin}_{bandmax}_{nperband*len(range_bands)}.hdf5"])
    odd_signal_filenames = os.path.join(*[path,"odd",brange,signal_out_snr,f"freq_{bandmin}_{bandmax}_{nperband*len(range_bands)}.hdf5"])

    for fname, temp_data in zip([even_noise_filenames, even_signal_filenames, odd_noise_filenames, odd_signal_filenames], [even_noise_save_outs, even_signal_save_outs, odd_noise_save_outs, odd_signal_save_outs]):
        if not os.path.isdir(os.path.dirname(fname)):
            os.makedirs(os.path.dirname(fname))
        print(fname, save_options)
        with h5py.File(fname,"w") as f:
            for key in save_options:
                if key == "pars":
                    # pars are stored as a dictionary so have to be saved differently to hdf5
                    parkeys = list(temp_data[key][0].keys())
                    f.create_dataset("parnames", data = parkeys)
                    f.create_dataset("pars", data = np.array([[td[key] for key in parkeys] for td in temp_data[key]]))
                else:
                    try:
                        f.create_dataset(key, shape=np.shape(temp_data[key]), data=np.array(temp_data[key]))
                    except ValueError as e:
                        print(key)
                        print(len(temp_data[key]))
                        lengths = [temp_data[key][i] for i in range(len(temp_data[key]))]
                        print(len(temp_data[key][0]))
                        print(max(lengths), min(lengths))
                        #print(np.shape(temp_data[key]))
                        #print(np.array(temp_data[key]).dtype)
                        #print(np.array(temp_data[key]).shape)
                        print(e)
                        raise Exception("Error on hdf5 save array")

    end_time = time.time()
    date = datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")
    with open(os.path.join(path,"timing.txt"),"a") as f:
        f.write("time-for-gen    {}    {} \n".format(end_time-start_time,date))


def run_and_inj(fmin,width,outpath,resize_image=False,snrrange = (50,150),av_sh=None,nsft=22538,tstart=931052949,tsft=1800.,test = False,save_options=None, stride = 1, degfree = 96, brange = ""):
    
    fmax = np.round(fmin + width,2)

    sig = cw.GenerateSignal()
    sig.alpha = 0
    sig.delta = 0
    sig.cosi = 0
    sig.phi0 = 0
    sig.psi = 0
    sig.f = [0,0]
    sig.earth_ephem = "/home/joseph.bayley/repositories/lalsuite/lalpulsar/lib/earth00-40-DE430.dat.gz"
    sig.sun_ephem = "/home/joseph.bayley/repositories/lalsuite/lalpulsar/lib/sun00-40-DE430.dat.gz"
    sig.snr = 0
    sig.fmin = fmin
    sig.fmax = fmax
    sig.tref = tstart
    
    Sn = {"H1":av_sh[0],"L1":av_sh[1]}
    if test:
        out_snr_noise = "./"
    else:
        out_snr_noise = "snr_0.0_0.0"
    data = sig.get_spectrogram(tstart = tstart, nsft = nsft,tref=tstart,tsft=tsft,fmin=fmin,fmax=fmax,snr=0,dets= ["H1","L1"],Sn=Sn)
    data.sum_sfts()
    data.downsamp_frequency(stride=stride,data_type="summed_norm_sft_power",remove_original=False)
    if np.shape(data.L1.downsamp_summed_norm_sft_power) != np.shape(data.H1.downsamp_summed_norm_sft_power):
        print(fmin,np.shape(data.L1.downsamp_summed_norm_sft_power),np.shape(data.H1.downsamp_summed_norm_sft_power))
        sys.exit()
    noise_outputs = cnn_data_gen.return_outputs(out_snr=out_snr_noise,datah=data.H1.downsamp_summed_norm_sft_power,datal=data.L1.downsamp_summed_norm_sft_power, fmin=fmin,fmax=fmax,snr=0,depth=0,h0=0, resize_image=resize_image, save_options = save_options, epochs = data.H1.summed_epochs, degfree = degfree, brange = brange)
    noise_outputs["pars"] = {"snr":0}
    noise_outputs["pars"]["tref"] = tstart
    noise_outputs["pars"]["snr"] = 0
    noise_outputs["pars"]["av_sh"] = np.nanmedian(av_sh)
    noise_outputs["pars"]["width"] = width
    noise_outputs["pars"]["fmin"] = data.H1.fmin
    noise_outputs["pars"]["fmax"] = data.H1.fmax

    del data, sig

    # inject signal into band

    sig = cw.GenerateSignal()
    
    snrstart,snrend = snrrange[0],snrrange[1]
            
    sigfreq = np.random.uniform(0,1)*(width)*0.5 + fmin + (width)*0.25
        
    snr = int(np.random.rand(1)*(snrend-snrstart) + snrstart)
        
    # freq, spindown^n, alpha, sin(delta), phi0, psi, cosi
    #params_low =  [fmin + 0.25*width,       -1e-9 ,0      ,-1,0      ,0      ,-1.0]
    #params_high = [fmin + width-0.25*width, -1e-16,2*np.pi, 1,2*np.pi,np.pi/2,1.0]
    params_low =  [fmin,       -1e-9 ,0      ,-1,0      ,0      ,-1.0]
    params_high = [fmin + width, -1e-16,2*np.pi, 1,2*np.pi,np.pi/2,1.0]

    data_list_signal = []
    data_list_noise = []
    nsft = int(nsft)
    param_list = ["f","fd","alpha","sindelta","phi0","psi","cosi"]
    pars = {}
    for j in np.linspace(0,len(param_list)-1,len(param_list)).astype(int):
        pars[param_list[j]] = params_low[j] + np.random.uniform(0,1)*(params_high[j] - params_low[j])

    #pars["fd"] = -10**(pars["logfd"])
    #pars["fd"] = np.random.uniform(-1,1)*10**pars["fd"]

    sig.alpha = pars["alpha"]
    sig.delta = np.arcsin(pars["sindelta"])
    sig.cosi = pars["cosi"]
    sig.phi0 = pars["phi0"]
    sig.psi = pars["psi"]
    sig.f = [pars["f"],pars["fd"]]
    sig.earth_ephem = "/home/joseph.bayley/repositories/lalsuite/lalpulsar/lib/earth00-40-DE430.dat.gz"
    sig.sun_ephem = "/home/joseph.bayley/repositories/lalsuite/lalpulsar/lib/sun00-40-DE430.dat.gz"
    sig.snr = snr
    sig.fmin = fmin
    sig.fmax = fmax
    sig.tref = tstart

    epochs = []

    print("in minmaxf", fmin, fmax)  
    data = sig.get_spectrogram(tstart = tstart, nsft = nsft,tref=tstart,tsft=tsft,fmin=fmin,fmax=fmax,snr=snr,dets= ["H1","L1"],Sn=Sn)
    data.sum_sfts()
    data.downsamp_frequency(stride=stride,data_type="summed_norm_sft_power",remove_original=False)
    if np.shape(data.L1.downsamp_summed_norm_sft_power) != np.shape(data.H1.downsamp_summed_norm_sft_power):
        print(np.shape(data.L1.downsamp_summed_norm_sft_power),np.shape(data.H1.downsamp_summed_norm_sft_power))
        sys.exit()

    print("freqs",fmin,fmax, data.H1.frequencies[0], data.H1.frequencies[-1] + 1./data.tsft)
    epochs = data.H1.summed_epochs

    h0 = np.sqrt(np.nanmedian(av_sh))/data.depth

    inj_track = (sig.get_pulsar_path(epochs=epochs,det="H1") - sig.fmin)*data.tsft 

    pars["tref"] = tstart
    pars["snr"] = snr
    pars["av_sh"] = np.nanmedian(av_sh)
    pars["h0"] = h0
    pars["depth"] = data.depth
    pars["width"] = width
    pars["fmin"] = data.H1.fmin
    pars["fmax"] = data.H1.fmax
    if test:
        out_snr_inj = "./"
    else:
        out_snr_inj = "snr_{}_{}".format(snrstart,snrend)

    signal_outputs = cnn_data_gen.return_outputs(out_snr=out_snr_inj,datah=data.H1.downsamp_summed_norm_sft_power,datal=data.L1.downsamp_summed_norm_sft_power, fmin=fmin,fmax=fmax,snr=snr, depth=data.depth,h0=h0,resize_image=resize_image,pars=pars,inj_track=inj_track,epochs=epochs, save_options=save_options, degfree = degfree, brange = brange)

    return noise_outputs, signal_outputs



import numpy as np
#from pycbc.frame import read_frame as read_frame_pycbc
from .sft import SFT

class TimeSeries:
    
    def __init__(self):
        """ Class to contain timeseries"""

        self.timeseries = None
        self.delta_t = None
        self.sample_frequency = None
        self.start_time = None
        self.end_time = None


    def LAL_sfts_from_timeseries(self):

        self.timeseries = self.timeseries.astype("float64") # make real8type
        self.length = len(self.timeseries)
        vect = lal.CreateREAL8Vector(self.length)
        vect.data = self.timeseries

        real8time = lal.CreateREAL8TimeSeries("", self.start_time, 0, self.delta_t, lal.SecondUnit, self.length)
        real8time.data = vect

        # !!!! NOT WORKING
        self.epochs = None
        lalpulsar.MakeSFTsFromREAL8TimeSeries(r8, self.epochs,"hann",0)
        
    def sfts_from_timeseries(self, tsft, window=None, fmin=None, fmax=None,real=False,tstart=None, overlap = 0):
        """
        Convert a timeseries array into a set of complex ffts or real fft
        Parameters
        ---------------------
        tsft: float
            length of an sft in seconds
        window: string (optional)
            type of window to use (only hanning right now will update) (optional)
        fmin: float
            lower frequency
        fmax: float
            upper frequency
        real: bool optional (True)
            return real part of fft if true
        tstart: int
            start time of fft
        overlap: float (optional)
            length of overlap as a fraction, i.e. 0.9 is a 90% overlap, default 0
        Returns
        ------------------
        sft: SFT
            return sft which can then be written to file etc
        
        """
        
        # initialise sft object
        sft = SFT()
        
        # define the sampling frequency as 1/dt if not defined already
        if self.sample_frequency is None:
            self.sample_frequency = 1./self.delta_t

        if self.delta_t is None:
            self.delta_t = 1./self.sample_frequency
            
        # define the frequency bin size for the sft
        sft_df = 1./tsft
            
        # set the bin index for fmin and fmax
        if fmin is None:
            fmin_index = int(0)
        else:
            fmin_index = int(fmin*tsft)
        if fmax is None:
            fmax_index = int(self.sample_frequency*tsft + 1)
        else:
            fmax_index = int(fmax*tsft + 1)
            
        # define the start time if not defined already
        if tstart is not None:
            if self.start_time is None:
                self.start_time = tstart
        elif tstart is None:
            if self.start_time is not None:
                tstart = self.start_time
            else:
                tstart = 0
        # set the detector name to none if it is not defined
        if not hasattr(self,"det_name"):
            self.det_name = "none"

        # set the length of each segment in time, and the length of the fft = N/2
        segment_length_index = int(tsft*self.sample_frequency)
        fft_length_index = int(segment_length_index/2)

        # split the time series up into segments of length segment_length_index overlapping by overlap
        # segments which re not the correct length are ignored, i.e. at the end of the data the segment will be cut off
        time_segments = []
        segment_number = np.floor(float(len(self.timeseries))/(segment_length_index*(1-overlap))).astype(int)
        for i in range(segment_number):
            temp_seg = np.array(self.timeseries[int(i*segment_length_index*(1-overlap)):int(int(i*segment_length_index*(1-overlap)) + segment_length_index)])
            if len(temp_seg) == segment_length_index:
                time_segments.append(temp_seg)
            else:
                print("Segment not correct length, omitting SFT no {}".format(i))
            del temp_seg

        
        tlen = int(len(time_segments))
        flen = int(len(time_segments[0]))
    
        
        # set up arrays for ffts if ral and complex
        if real:
            ffts = np.zeros((tlen, int(fft_length_index + 1)))
            fft_freqs = np.zeros((tlen, int(fft_length_index + 1)))
        elif not real:
            ffts = np.zeros((tlen, int(fft_length_index + 1)), dtype=np.complex_)
            fft_freqs = np.zeros((tlen, int(fft_length_index + 1 )))

        # create each sft
        epochs = np.zeros(tlen)
        for t,seg in enumerate(time_segments):
            # set window function, default hanning, need to add more options
            if window == "hann" or window == "hanning":
                seg_wind = np.hanning(segment_length_index) * seg
            elif window is None:
                seg_wind = seg
            elif type(window) in [list,np.array,np.ndarray]:
                # use input array as window function
                if len(window) != len(seg):
                    raise Exception("Window must be same length as the segment")
                else:
                    seg_wind = window * seg
            if real:
                epochs[t] = tstart + t*tsft
                fft_calc = np.fft.rfft(seg_wind)
                ffts[t] = fft_calc#[:int(len(fft_calc/2) )]
                fft_freqs[t] = np.fft.rfftfreq(len(ffts[t]), self.delta_t)
            else:
                epochs[t] = tstart + t*tsft
                fft_calc = np.fft.fft(seg_wind)
                # take first half of fft
                ffts[t] = fft_calc[:int(len(fft_calc)/2 + 1)]
                # take absolute of fft freqs as including the nyquist freq is negative 
                fft_freqs[t] = np.abs(np.fft.fftfreq(len(fft_calc), self.delta_t)[:int(len(fft_calc)/2 + 1)])

        # set paramters of the sft
        sft.frequencies = fft_freqs[:,fmin_index:fmax_index]
        sft.fmin = sft.frequencies.min()
        sft.fmax = sft.frequencies.max()
        sft.delta_f = sft_df
        sft.sample_frequency = self.sample_frequency
        sft.tsft = tsft
        sft.sft = ffts[:,fmin_index:fmax_index]
        sft.epochs = epochs
        sft.det_name = self.det_name
        sft.nsft = segment_number
        return sft
    
    def write_sft_from_timeseries(self,output_path,tsft,window=None,fmin=None,fmax=None,narrowband = False):
        """
        write a gwf file to an sft
        args
        ----------
        output_path: string
            path to output file if narrowband, path to output dir if not narrowband
        tsft: float
            length of each sft
        window: string or array {optional:"hann"}
            window name, or array same width as data segment
        fmin: float {optional: 0}
            minimum frequency
        fmax: float {optional: }
            maximum frequency
        narrowband: bool {optional: False}
            either write to a narrowband combined set of sfts (True), or write different sft file for each sft (False)

        """
        
        sft = self.sfts_from_timeseries(tsft, window=window, fmin=fmin, fmax=fmax,real=False)
        sft.write_sft_files(output_path,narrowband=narrowband)
        
class LoadTimeSeries(TimeSeries):
    
    def __init__(self,filename,channel_name,start_time=None,end_time=None):
        """
        Load timeseries as an array from gwf file, uses pycbcs readframe
        Args
        -----------
        filename: string
            gwf filename
        channel_name: string:
            data channel_name
        start_time: int (optional: default - None)
            start time where you want to read the gwf file from, default is to read whole file
        end_time: int (optional: default - None)
            dn time where you want to read the gwf file to, default is to read whole file
        """
        self.read_frame(filename=filename,channel_name=channel_name,start_time=start_time,end_time=end_time)
    
    def read_frame(self,filename,channel_name,start_time=None,end_time=None):
        
        fr = read_frame_pycbc(filename,channels = channel_name,start_time=start_time,end_time=end_time)
        
        self.timeseries = np.array(fr.data)
        self.delta_t = float(fr.delta_t)
        self.delta_f = float(fr.delta_f)
        self.sample_frequency = float(fr.sample_rate)
        self.start_time = float(fr.start_time)
        self.end_time = float(fr.end_time)
        self.duration = float(fr.duration)
        
        # temporary want to get detectro from gwf file
        self.det_name = channel_name.split(":")[0]
        




import numpy as np
from scipy.signal import spectrogram
import lal
import lalpulsar
import pycbc
try:
    from pylal import Fr
except:
    print("No module Pylal installed, some funtions will fail")#from pylal import Fr



class MakeSFTs(object):

    def __init__(self):
        pass

    def get_spectrogram(self,frame_filename, channel_name, fft_length, overlap=None, fmin=None, fmax=None):
        """
        A small script to make a spectrogram of a gwf frame file using scipys spectrogram
        Parameters
        --------
        frame_filename: string
            path to frame file
        channel_name: string
            name of channel
        fft_length: int
            length of the fft in seconds 
        overlap: float (optional)
            what percentage are the SFTs overlapping by
        window: tuple (optional)
            which window type (inputs from scipy.signal.spectrogram)
        fmin: float (optional)
            lower frequency of sfts
        fmax: float (optional)
            upper frequency of sfts        
        """
        # read strain fro frame file
        # need to change to read in multiple frame files
        self.strain, self.gpsstart, self.ts = self.read_frame(frame_filename, channel_name)
        self.fs = 1./self.ts
        #create spectrogram 
        spect = self.make_sft(self.strain, self.fs, fft_length, overlap=overlap, fmin=fmin,fmax=fmax)
        self.spectrogram = spect[2]
        self.frequencies = spect[0]
        self.times = spect[1]

        if meandivide:
            self.mean_divide()

    def make_sft_from_gwf(self,output_path, frame_filename, channel_name, fft_length, detector_name, window = None, fmin=None, fmax = None):
        """
        take input of gwf frame files and output set of sfts of length fft_length
        Parameters
        ------------------
        output_path: string
            path to save sfts to
        frame_filename: string or list
            string of frame filename or list of filenames
        channel_name: string
            channel to load within the frame file
        fft_length: float
            length of the fft in seconds
        detector_name: string
            string of detector name, i.e. 'H1', 'L1'
        window: string (optional)
            which window type to use (default None) (only hanning available right now)
        fmin: float
            lower frequency in Hz
        fmax: float
            upper frequency in Hz
        """
        # read strain fro frame file
        # need to change to read in multiple frame files
        self.strain, self.gpsStart, self.ts = self.read_frame(frame_filename, channel_name)
        self.fs = 1./self.ts

        if fmin is None:
            fmin_index = int(0)
            fmin = 0
        else:
            fmin_index = int(fmin*fft_length)
            fmin = fmin
        if fmax is None:
            fmax_index = int(self.fs*fft_length)
            fmax = fft_length/2.
        else:
            fmax_index = int(fmax*fft_length)
            fmax = fft_length/2.

        #create ffts
        fft_freqs, ffts = self.sfts_from_timeseries(self.strain, self.fs, fft_length, window=window, fmin=fmin_index, fmax=fmax_index)

        # write ffts
        self.sft_file_from_array(output_path, ffts, self.fs, fft_length, detector_name, self.gpsStart, fmin,fmax)



    def read_frame(self,filename, strain_name="H1:GDS-CALIB_STRAIN", readstrain=True):
        """
        Read frame files
        Parameters
        ---------
        filename: string
            path to frame file
        channel_name: string
            name of channel
        Returns
        ---------
        strain: array
            array of string values
        gpsStart: float
            start time in gps time
        ts: float
            time between samples
        """
        if type(filename) is str:
            sd = Fr.frgetvect(filename, strain_name)    
            strain = sd[0]
            gpsStart = sd[1] 
            ts = sd[3][0]
        elif type(filename) == list:
            strain = []
            gpsStart = 0
            ts = 0
            for f in filename:
                # need to add in condition to make sure all same samplig etc and check all in order, i.e can join them up
                # not finished
                sd = Fr.frgetvect(filename, strain_name)
                strain.append(sd[0])
                gpsStart = sf[1]
                ts = sd[3][0]

        return strain, gpsStart, ts


    def mean_divide(self):
        """
        Divide the spectrogram by the mean in time
        very very basic attempt at line removal
        """
        meansp = np.mean(self.spectrogram.T,axis=0)
        for i in range(len(self.spectrogram.T)):
            self.spectrogram.T[i] = self.spectrogram.T[i]/meansp 

    def sfts_from_timeseries(self,timeseries, fs, fft_length, window=None, fmin=None, fmax=None,real=True):
        """
        Convert a timeseries array into a set of complex ffts
        Parameters
        ---------------------
        timeseries: array
            timeseries to be converted
        fs: float
            frequency of sampling in Hz
        fft_length: float
            segment length of times series to create fft from
        window: string (optional)
            type of window to use (only hanning right now will update) (optional)
        fmin: float
            lower frequency
        fmax: float
            upper frequency
        
        Returns
        ------------------
        fft_freqs: array
            2d array of frequencies for each segment
        ffts: array
            2d array of the complex ffts for each segment
        
        """

        if fmin is None:
            fmin_index = int(0)
        else:
            fmin_index = int(fmin*fft_length)
        if fmax is None:
            fmax_index = int(fs*fft_length)
        else:
            fmax_index = int(fmax*fft_length)

        fft_length_index = int(fft_length*fs)

        time_segments = []
        segment_number = np.floor(float(len(timeseries))/fft_length_index).astype(int)

        for i in range(segment_number):
            time_segments.append(np.array(timeseries[i*fft_length_index:fft_length_index*(i+1)]))

        if real:
            ffts = np.zeros((len(time_segments),len(time_segments[0])/2))
            fft_freqs = np.zeros((len(time_segments),len(time_segments[0])))#/2 + 1))
        elif not real:
            ffts = np.zeros((len(time_segments),len(time_segments[0])/2),dtype=np.complex_)
            fft_freqs = np.zeros((len(time_segments),len(time_segments[0])))#/2 + 1))

        for t,seg in enumerate(time_segments):
            if window == "hann":
                seg_wind = np.hanning(fft_length_index) * seg
            elif window is None:
                seg_wind = seg
            if real:
                fft_calc = np.fft.rfft(seg_wind)
                ffts[t] = fft_calc[:int(len(fft_calc)/2 )]
                fft_freqs[t] = np.fft.rfftfreq(fft_length_index, 1./fs)
            else:
                fft_calc = np.fft.fft(seg_wind)
                ffts[t] = fft_calc[:int(len(fft_calc)/2 )]
                fft_freqs[t] = np.fft.fftfreq(fft_length_index, 1./fs)

        fft_freqs_slice = fft_freqs[:,fmin_index:fmax_index]
        ffts_slice = ffts[:,fmin_index:fmax_index]
        return fft_freqs_slice, ffts_slice

    def sft_file_from_array(self, output_path, ffts, fs, fft_length, detector_name, gpsStart, fmin, fmax):
        """
        Save and array of ffts to a set of sft files ()
        Parameters
        ---------------------
        output_path: string
            path where sfts will be saved
        ffts: array
            array of complex ffts for each segment in ascending time order
        fs: float
            frequency of sampling in Hz
        fft_length: float
            segment length of times series to create fft from
        detector_name: string
            string for detector name i.e. L1,H1 etc
        gpsStart: int
            the start gps time
        f0: float
            start frequency

        """

        fft_length_index = int(fft_length*fs)

        for t,fft in enumerate(ffts):
            sft_temp = lal.CreateCOMPLEX8FrequencySeries(detector_name, lal.LIGOTimeGPS(gpsStart), fmin, 1./fft_length, lal.SecondUnit, int(fft_length_index/2.))

            sft_temp.data.data[:] = fft
            filename_prefix = "{}/{}_DT{}_T{}_F{}_{}.sft".format(output_path, detector_name,fft_length, gpsStart + t*fft_length , fmin,fmax)
            lalpulsar.WriteSFT2file(sft_temp, filename_prefix, None)

        return None
            
            



import torch
import torch.nn as nn
import numpy as np
import time
from .truncated_gauss import TruncatedNormal

class CVAE(nn.Module):

    def __init__(self, input_dim, par_dim, latent_dim, fc_layers = [], conv_layers = [], stride = 1, device="cpu", dropout=0.1, fdim=0, inchannels = 1, dist_type = "gaussian"):
        """
        args
        ----------
        input_dim: int
            length of input time series
        par_dim: int
            size of hidden layers in all encoders
        latent_dim: int
            number of variables in latent space
        par_dim: int
            number of parameter to estimat
        fc_layers: list
            [num_neurons_layer_1, num_neurons_layer_2,....]
        conv_layers: list
            [(num_filters, conv_size, dilation, maxpool size), (...), (...)]
        """
        super().__init__()
        # define useful variables of network
        self.device = device
        self.input_dim = input_dim
        self.par_dim = par_dim
        self.output_dim = self.input_dim + self.par_dim
        self.gauss_par_dim = self.par_dim - 1
        self.latent_dim = latent_dim
        self.inchannels = inchannels
        self.dist_type = dist_type

        # convolutional parts
        self.fc_layers = fc_layers
        #self.fc_layers_decoder = np.array(0.5*np.array(fc_layers[:3])).astype(int)
        self.conv_layers = conv_layers
        self.num_conv = len(self.conv_layers)
        self.stride = stride

        self.trunc_mins = torch.Tensor([0,0,0,0]).to(self.device)
        self.trunc_maxs = torch.Tensor([1,1,1,1]).to(self.device)
        self.activation = nn.ReLU()
        self.tanh = nn.Tanh()
        self.sigmoid = nn.Sigmoid()
        self.drop = nn.Dropout(p=dropout)

        self.small_const = 1e-6
        
        self.fdim = fdim
        
        # encoder r1(z|y) 
        self.rencoder_conv, self.rencoder_lin, _ctr = self.create_network("r", self.input_dim, self.latent_dim, append_dim=self.fdim, fc_layers = self.fc_layers, conv_layers = self.conv_layers)

        # conv layers not used for next two networks, they both use the same rencoder
        # encoder q(z|x, y) 
        qencoder_conv, self.qencoder_lin, self.conv_track_network = self.create_network("q", self.input_dim, self.latent_dim, append_dim=self.par_dim + self.fdim, fc_layers = self.fc_layers, conv_layers = self.conv_layers, track_conv = True)

        # decoder r2(x|z, y) 
        decoder_conv, self.decoder_lin, _ctr = self.create_network("d", self.input_dim, self.par_dim, append_dim=self.latent_dim + self.fdim, fc_layers = self.fc_layers, conv_layers = self.conv_layers, meansize = self.output_dim)
        
    def create_network(self, name, input_dim, output_dim, append_dim=0, mean=True, variance=True, weight = False,fc_layers=[], conv_layers=[], meansize = None, track_conv = False):
        """ Generate arbritrary network, with convolutional layers or not
        args
        ------
        name: str
            name of network
        input_dim: int
            size of input to network
        output_dim: int
            size of output of network
        append_dim: int (optional)
            number of neurons to append to fully connected layers after convolutional layers
        mean: bool (optional) 
            if True adds extra output layer for means 
        variance : bool (optional)
            if True adds extra layer of outputs for variances
        fc_layers: list
            list of fully connected layers, format: [64,64,...] = [num_neurons_layer1, num_neurons_layer2, .....]
        conv_layers: list
            list of convolutional layers, format:[(8,8,2,1), (8,8,2,1)] = [(num_filt1, conv_size1, max_pool_size1, dilation1), (num_filt2, conv_size2, max_pool_size2, dilation2)] 
        """
        conv_network = nn.Sequential() # initialise networks
        conv_track_network = nn.Sequential()
        lin_network = nn.Sequential()
        layer_out_sizes = []
        num_conv = len(conv_layers)
        num_fc = len(fc_layers)
        inchannels = self.inchannels
        insize = self.input_dim
        # add convolutional layers
        for i in range(num_conv):
            padding = int(conv_layers[i][1]/2.) # padding half width
            maxpool = nn.MaxPool1d(conv_layers[i][3]) #define max pooling for this layer
            # add convolutional/activation/maxpooling layers
            conv_network.add_module("r_conv{}".format(i),module = nn.Conv1d(inchannels, conv_layers[i][0], conv_layers[i][1], stride = self.stride,padding=padding, dilation = conv_layers[i][2]))    
            #conv_network.add_module("batch_norm_conv{}".format(i), module = nn.BatchNorm1d(conv_layers[i][0]))
            conv_network.add_module("act_r_conv{}".format(i), module = self.activation)
            conv_network.add_module("pool_r_conv{}".format(i),module = maxpool)
            # define the output size of the layer
            outsize = int(self.conv_out_size(insize, padding, conv_layers[i][2], conv_layers[i][1], self.stride)/conv_layers[i][3]) # output of one filter
            layer_out_sizes.append((conv_layers[i][0],outsize))
            insize = outsize
            inchannels = conv_layers[i][0]

        if track_conv:
            layer_track_out_sizes = []
            num_conv_track = 1
            inchannels_track = 1
            insize_track = self.input_dim
            conv_layers_track = [(4,3,1,16),]

            for i in range(num_conv_track):
                padding = int(conv_layers_track[i][1]/2.) # padding half width
                maxpool_track = nn.MaxPool1d(conv_layers_track[i][3]) #define max pooling for this layer
                # add convolutional/activation/maxpooling layers
                conv_track_network.add_module("r_conv{}".format(i),module = nn.Conv1d(inchannels_track, conv_layers_track[i][0], conv_layers_track[i][1], stride = self.stride,padding=padding, dilation = conv_layers_track[i][2]))    
                conv_track_network.add_module("act_r_conv{}".format(i), module = self.activation)
                #conv_network.add_module("batch_norm_conv{}".format(i), module = nn.BatchNorm1d(conv_layers[i][0]))
                conv_track_network.add_module("pool_r_conv{}".format(i),module = maxpool_track)
                # define the output size of the layer
                outsize_track = int(self.conv_out_size(insize_track, padding, conv_layers_track[i][2], conv_layers_track[i][1], self.stride)/conv_layers_track[i][3]) # output of one filter
                layer_track_out_sizes.append((conv_layers_track[i][0],outsize_track))
                insize_track = outsize_track
                inchannels_track = conv_layers_track[i][0]

            track_append_size = np.prod(layer_track_out_sizes[-1])
            append_dim += track_append_size

        # define the input size to fully connected layer
        lin_input_size = np.prod(layer_out_sizes[-1]) if num_conv > 0 else self.input_dim
        if append_dim:
            lin_input_size += append_dim
        
        layer_size = int(lin_input_size)
        # hidden layers
        for i in range(num_fc):
            lin_network.add_module("r_lin{}".format(i),module=nn.Linear(layer_size, fc_layers[i]))
            #lin_network.add_module("batch_norm_lin{}".format(i), module = nn.BatchNorm1d(fc_layers[i]))
            lin_network.add_module("r_drop{}".format(i),module=self.drop)
            lin_network.add_module("act_r_lin{}".format(i),module=self.activation)
            layer_size = fc_layers[i]
        # output mean and variance of gaussian with size of latent space

        if mean:
            if meansize is None:
                meansize = output_dim
            setattr(self,"mu_{}".format(name[0]),nn.Linear(layer_size, meansize))
        if variance:
            setattr(self,"log_var_{}".format(name[0]),nn.Linear(layer_size, output_dim))
        if self.dist_type is not "gaussian":
            if weight:
                setattr(self,"weight_{}".format(name[0]),nn.Linear(layer_size, 1))

        return conv_network, lin_network, conv_track_network

    def conv_out_size(self, in_dim, padding, dilation, kernel, stride):
        """ Get output size of a convolutional layer (or one filter from that layer)"""
        return int((in_dim + 2*padding - dilation*(kernel-1)-1)/stride + 1)
        
    def encode_r(self,y, freqs):
        """ encoder r1(z|y) , takes in observation y"""
        conv = self.rencoder_conv(torch.reshape(y, (y.size(0), self.inchannels, self.input_dim))) if self.num_conv > 0 else y
        lin_in = torch.flatten(conv,start_dim=1)
        if freqs is not None:
            lin_in = torch.cat([lin_in, torch.unsqueeze(freqs, 1)],1)
        lin = self.rencoder_lin(lin_in)
        z_mu = self.mu_r(lin) # latent means
        z_log_var = self.log_var_r(lin) + self.small_const # latent variances
        return z_mu, z_log_var
    
    def encode_q(self,y,par, freqs):
        """ encoder q(z|x, y) , takes in observation y and paramters par (x)"""
        conv = self.rencoder_conv(torch.reshape(y, (y.size(0), self.inchannels, self.input_dim))) if self.num_conv > 0 else y
        conv_track = self.conv_track_network(torch.reshape(par[:,4:], (par.size(0), 1, self.input_dim)))
        par = torch.cat([torch.flatten(conv_track,start_dim=1), par[:,:4]],1)
        lin_in = torch.cat([torch.flatten(conv,start_dim=1), par],1)
        if freqs is not None:
            lin_in = torch.cat([lin_in, torch.unsqueeze(freqs, 1)],1)
        lin = self.qencoder_lin(lin_in)
        z_mu = self.mu_q(lin)  # latent means
        z_log_var = self.log_var_q(lin) + self.small_const # latent vairances
        return z_mu, z_log_var
    
    def decode(self, z, y, freqs):
        """ decoder r2(x|z, y) , takes in observation y and latent paramters z"""
        conv = self.rencoder_conv(torch.reshape(y, (y.size(0), self.inchannels, self.input_dim))) if self.num_conv > 0 else y
        lin_in = torch.cat([torch.flatten(conv,start_dim=1),z],1) 
        if freqs is not None:
            lin_in = torch.cat([lin_in, torch.unsqueeze(freqs, 1)],1)
        lin = self.decoder_lin(lin_in)
        par_mu = self.mu_d(lin) # parameter means
        par_mu[:,:4] = self.sigmoid(par_mu[:,:4])
        #par_mu[:,:3] = self.sigmoid(par_mu[:,:3])
        par_log_var = self.log_var_d(lin) + self.small_const # parameter variances
        return par_mu, par_log_var

    def gauss_sample(self, mean, log_var, num_batch, dim):
        """ Sample trom a gaussian with given mean and log variance 
        (takes in a number (dim) of means and variances, and samples num_batch times)"""
        std = torch.exp(0.5 * (log_var))
        eps = torch.randn([num_batch, dim]).to(self.device)
        sample = torch.add(torch.mul(eps,std),mean)
        return sample

    def trunc_gauss_sample(self, mu, log_var, ramp = 1.0):
        """Gaussian log-likelihood """
        sigma = torch.sqrt(torch.exp(log_var))
        #dist = TruncatedNormal(mu, sigma, self.trunc_mins - 10 + ramp*10, self.trunc_maxs + 10 - ramp*10)
        dist = TruncatedNormal(mu, sigma, 0, 1)
        return dist.sample()

    def log_likelihood_trunc_gauss(self,par, mu, log_var, ramp = 1.0):
        """Gaussian log-likelihood """
        sigma = torch.sqrt(torch.exp(log_var))
        try:
            #dist = TruncatedNormal(mu, sigma, a=self.trunc_mins - 10 + ramp*10, b=self.trunc_maxs + 10 - ramp*10)
            dist = TruncatedNormal(mu, sigma, a=0, b=1)
        except:
            #print((self.trunc_mins - 10 + ramp*10 - mu)/sigma, (self.trunc_maxs + 10 - ramp*10 - mu)/sigma, ramp*10
            print(mu)
        return dist.log_prob(par)

    def log_likelihood_cat(self, par, logits):
        """Gaussian log-likelihood """ 
        cat = torch.distributions.binomial.Binomial(total_count = 1, logits = logits)
        logliks = cat.log_prob(par)
        # takes the mean of the likelihoods in the parameter space
        return logliks

    def cat_samp(self, logits):
        cat = torch.distributions.binomial.Binomial(total_count = 1, logits = logits)
        return cat.sample()
    
    def KL_gauss(self,mu_r,log_var_r,mu_q,log_var_q):
        """Gaussian KL divergence between two distributions"""
        sigma_q = torch.exp(0.5 * (log_var_q))
        sigma_r = torch.exp(0.5 * (log_var_r))
        t2 = torch.log(sigma_r/sigma_q)
        t3 = (torch.square(mu_q - mu_r) + torch.square(sigma_q))/(2*torch.square(sigma_r))
        # take sum of KL divergences in the latent space
        kl_loss = torch.sum(t2 + t3 - 0.5,dim=1)
        return kl_loss


    def forward(self, y, par, freqs):
        """forward pass for training"""
        batch_size = y.size(0) # set the batch size
        # encode data into latent space
        mu_r, log_var_r = self.encode_r(y, freqs) # encode r1(z|y)
        mu_q, log_var_q = self.encode_q(y, par, freqs) # encode q(z|x, y)
        
        # sample z from gaussian with mean and variance from q(z|x, y)
        z_sample = self.gauss_sample(mu_q, log_var_q, batch_size, self.latent_dim)
        #z_sample = self.lorentz_sample(mu_q, log_var_q)[0]
        # get the mean and variance in parameter space from decoder
        if self.dist_type == "studentT":
            mu_par, log_var_par, weight_par = self.decode(z_sample,y, freqs) # decode r2(x|z, y)                                                                              
            return mu_par, log_var_par, weight_par, mu_q, log_var_q, mu_r, log_var_r
        elif self.dist_type == "gaussian":
            mu_par, log_var_par = self.decode(z_sample,y, freqs) # decode r2(x|z, y)                                                                              
            return mu_par, log_var_par, mu_q, log_var_q, mu_r, log_var_r

    def compute_loss(self, y, par, freqs, ramp):
        """
        Comput the loss function given input data 
        """
        mu_r, log_var_r = self.encode_r(y, freqs) # encode r1(z|y)     
        mu_q, log_var_q = self.encode_q(y, par, freqs) # encode q(z|x, y)  

        z_sample = self.gauss_sample(mu_q, log_var_q, mu_q.size(0), mu_q.size(1))

        mu_par, log_var_par = self.decode(z_sample,y, freqs) # decode r2(x|z, y)

        kl_loss = torch.mean(self.KL_gauss(mu_r, log_var_r, mu_q, log_var_q))
        
        gauss_loss = self.log_likelihood_trunc_gauss(par[:,:4], mu_par[:,:4], log_var_par, ramp)

        cat_loss = self.log_likelihood_cat(par[:,4:], mu_par[:,4:])

        recon_loss = -1.0*torch.mean(torch.sum(torch.cat([gauss_loss, cat_loss], axis = 1), axis = 1))
        
        return recon_loss, kl_loss

        
    
    def draw_samples(self, y, mu_r, log_var_r, num_samples, latent_dim, par_dim, return_latent = False, freqs = None):
        """ Draw samples from network for testing"""

        z_sample = self.gauss_sample(mu_r, log_var_r, num_samples, latent_dim)

        # input the latent space samples into decoder r2(x|z, y)  
        #ys = y.repeat(1,num_samples).view(-1, y.size(0)) # repeat data so same size as the z samples
        ys = y.repeat(num_samples,1,1).view(-1, y.size(0), y.size(1)) # repeat data so same size as the z samples
        freqs = freqs.repeat(num_samples).view(-1) # repeat data so same size as the z samples

        # sample parameter space from returned mean and variance 
        mu_par, log_var_par = self.decode(z_sample,ys, freqs) # decode r2(x|z, y) from z        
        samp = self.trunc_gauss_sample(mu_par[:,:4], log_var_par[:,:4])
        
        if np.any(samp[:, :3].cpu().numpy() > 1) or np.any(samp.cpu().numpy()[:,:3] < 0):
            print("samples outside range")
            g1ind = np.where(samp[:,:3].cpu().numpy() > 1)
            l0ind = np.where(samp[:,:3].cpu().numpy() < 0)
            print(samp[g1ind], mu_par[g1ind], log_var_par[g1ind])
            print(samp[l0ind], mu_par[l0ind], log_var_par[l0ind])
            sys.exit()
        # draw only one realisation for the tracks
        track_samp = self.cat_samp(mu_par[:,4:])

        if return_latent:
            return samp.cpu().numpy(), z_sample.cpu().numpy(), track_samp.cpu().numpy()
        else:
            return samp.cpu().numpy(), None, track_samp.cpu().numpy()
    
    def test_latent(self, y, freqs, par, num_samples):
        """generating samples when testing the network, returns latent samples as well (used during training to get latent samples)"""
        num_data = y.size(0)                                                                                                                                                     
        x_samples = []
        track_samples = []
        # encode the data into latent space with r1(z,y)          
        mu_r, log_var_r = self.encode_r(y, freqs) # encode r1(z|y) 
        mu_q, log_var_q = self.encode_q(y,par, freqs) # encode q(z|y) 
        # get the latent space samples
        zr_samples = []
        zq_samples = []
        for i in range(num_data):
            # sample from both r and q networks
            zr_sample = self.gauss_sample(mu_r[i], log_var_r[i], num_samples, self.latent_dim)
            zq_sample = self.gauss_sample(mu_q[i], log_var_q[i], num_samples, self.latent_dim)
            zr_samples.append(zr_sample.cpu().numpy())
            zq_samples.append(zq_sample.cpu().numpy())
            # input the latent space samples into decoder r2(x|z, y)  
            ys = y[i].repeat(num_samples,1,1).view(-1, y.size(1), y.size(2)) # repeat data so same size as the z samples
            freqss = freqs[i].repeat(num_samples).view(-1) # repeat data so same size as the z samples

            mu_par, log_var_par = self.decode(zr_sample,ys, freqss) # decode r2(x|z, y) from z        
            samp = self.trunc_gauss_sample(mu_par[:,:4], log_var_par[:,:4])
            track_samp = self.cat_samp(mu_par[:,4:])

            # add samples to list    
            x_samples.append(samp.cpu().numpy())
            track_samples.append(track_samp.cpu().numpy())
        return np.array(x_samples),np.array(zr_samples),np.array(zq_samples), track_samples
    
    
    def test(self, y, freqs, num_samples, transform_func=None, return_latent = False, par = None):
        """generating samples when testing the network 
        args
        --------
        model : pytorch model
            the input model to test
        y: Tensor
            Tensor of all observation data to generate samples from 
        num_samples: int
            number of samples to draw
        transform_func: function (optional)
            function which transforms the parameters into real parameter space
        return_latent: bool
            if true returns the samples in the latent space as well as output
        par: list (optional)
            parameter for each injection, used if returning the latent space samples (return_latent=True)
        """
        num_data = y.size(0)                                                                                                                                                     
        transformed_samples = []
        net_samples = []
        track_samples = []
        if return_latent:
            z_samples = []
            q_samples = []
        # encode the data into latent space with r1(z,y)          
        mu_r, log_var_r = self.encode_r(y, freqs) # encode r1(z|y) 
        if return_latent:
            mu_q, log_var_q = self.encode_q(y,par, freqs) # encode q(z|y) 
        # get the latent space samples for each input
        for i in range(num_data):
            print("index: {}".format(i))
            # generate initial samples
            t_net_samples, t_znet_samples, t_track_samples = self.draw_samples(y[i], mu_r[i], log_var_r[i], num_samples, self.latent_dim, self.par_dim, return_latent = return_latent, freqs=freqs[i])
            if return_latent:
                q_samples.append(self.gauss_sample(mu_q[i], log_var_q[i], num_samples, self.latent_dim).cpu().numpy())
            if transform_func is None:
                # if nans in samples then keep drawing samples until there are no Nans (for whan samples are outside prior)
                if np.any(np.isnan(t_net_samples)):
                    num_nans = np.inf
                    stime = time.time()
                    while num_nans > 0:
                        nan_locations = np.where(np.any(np.isnan(t_net_samples), axis=1))
                        num_nans = len(nan_locations[0])
                        if num_nans == 0: break
                        temp_new_net_samp, temp_new_z_sample, temp_track_samples = self.draw_samples(y[i], mu_r[i], log_var_r[i], num_nans, self.latent_dim, self.par_dim, return_latent = return_latent, freqs = freqs[i])
                        t_net_samples[nan_locations] = temp_new_net_samp
                        t_track_samples[nan_locations] = temp_track_samples
                        if return_latent:
                            t_znet_samples[nan_locations] = temp_new_z_sample

                        etime = time.time()
                        # if it still nans after 1 min cancel
                        if etime - stime > 0.5*60:
                            print("Failed to find samples within 3 mins")
                            num_nans = 0
                            break

                transformed_samples.append(t_net_samples)
                net_samples.append(t_net_samples)
                track_samples.append(t_track_samples)
                if return_latent:
                    z_samples.append(t_znet_samples)
                
            else:
                # transform all samples to new parameter space
                new_samples = transform_func(t_net_samples, i)
                # if transformed samples are outside prior (nan) then redraw nans until all real values
                if np.any(np.isnan(new_samples)):
                    stime = time.time()
                    num_nans = np.inf
                    while num_nans > 0:
                        nan_locations = np.where(np.any(np.isnan(new_samples), axis=1))
                        num_nans = len(nan_locations[0])
                        if num_nans == 0: break
                        #redraw samples at nan locations
                        temp_new_net_samples, temp_new_z_samples, temp_track_samples = self.draw_samples(y[i], mu_r[i], log_var_r[i], num_nans, self.latent_dim, self.par_dim, return_latent = return_latent, freqs = freqs[i])
                        transformed_newsamp = transform_func(temp_new_net_samples, i)
                        new_samples[nan_locations] = transformed_newsamp
                        t_net_samples[nan_locations] = temp_new_net_samples
                        t_track_samples[nan_locations] = temp_track_samples
                        if return_latent:
                            t_znet_samples[nan_locations] = temp_new_z_samples
                        etime = time.time()
                        # if it still nans after 1 min cancel
                        if etime - stime > 0.5*60:
                            print("Failed to find samples within 30s")
                            num_nans = 0
                            break
                    

                transformed_samples.append(new_samples)
                net_samples.append(t_net_samples)
                track_samples.append(t_track_samples)
                if return_latent:
                    z_samples.append(t_znet_samples)
        
        if return_latent:
            return np.array(transformed_samples), np.array(net_samples), np.array(z_samples), np.array(q_samples), np.array(track_samples)
        else:
            return np.array(transformed_samples), np.array(net_samples), np.array(track_samples)




from __future__ import print_function
import numpy as np
import matplotlib
try:
    import matplotlib.pyplot as plt
    import seaborn as sns
except:
    print("cannot import matplotlib")
import timeit
import scipy.stats as st
import os
import copy
import pickle
from scipy.signal import butter, filtfilt
from scipy.special import logsumexp


def day_track_sum(track):
    '''
    finds the track of a pulsar in the day average case
    args
    ------
    track: array
        path in 1800s sfts
    returns: array
        averages path over one day
    '''
    data_av1 = []
    for i in np.linspace(0,len(track)-49,len(track)-48)[::48]:
        av = np.mean(track[int(i):int(i+48)])
        data_av1.append(av)
    return np.array(data_av1)

def transition_matrix(par,log=True):
    '''
    generate a symmetric transition matrix for 0 memory and 1 detector
    args
    -------
    par: float
        ratio of two outer elements to the centre element of the transition matrix
    kwargs
    ------
    log: bool
        return the log of the transition matrix if True, just probabilities if false
    returns 
    -------
    tr: array
        3x1 array of transition probabilities 
    '''
    v1 = 1.
    v2 = v1/par
    tr = [v2,v1,v2]
    if log:
        return np.log(tr/np.sum(tr))
    if not log:
        return tr/np.sum(tr)

def tr_p_long(par,log=True,n=2):
    '''
    generate a symmetric transition matrix for 0 memory and 1 detector
    args
    -------
    par: float
        ratio of two outer elements to the centre element of the transition matrix
    kwargs
    ------
    log: bool
        return the log of the transition matrix if True, just probabilities if false
    returns 
    -------
    tr: array
        3x1 array of transition probabilities 
    '''
    v1 = 1.
    v2 = v1/par
    if n == 2:
        tr = [v2,v2,v1,v2,v2]
    if n == 3:
        tr = [v2,v2,v2,v1,v2,v2,v2]
    if log:
        return np.log(tr/np.sum(tr))
    if not log:
        return tr/np.sum(tr)

def transition_matrix_2d(par1,par2,par3,log=True):
    '''
    generate a symmetric transition matrix for 0 memory and 1 detector
    args
    -------
    par1: float
        ratio of left and rightjump elements to the centre jump element of the transition matrix
    par2: float
        ratio detector 1 being in left or right bin compared to reference
    par3: float
        ratio detector 2 being in left or right bin compared to reference
    kwargs
    ------
    log: bool
        return the log of the transition matrix if True, just probabilities if false
    returns 
    -------
    tr: array
        3x3 array of transition probabilities 
    '''
    j1 = 1.
    jlr = j1/par1
    plr1 = j1/par2
    plr2 = j1/par3
    tr_j= np.array([jlr,j1,jlr])
    tr_p1 = np.array([plr1,j1,plr1])
    tr_p2 = np.array([plr2,j1,plr2])
    tr_p = np.array([tr_p1[0]*tr_p2,tr_p1[1]*tr_p2,tr_p1[2]*tr_p2])
    tr = [tr_p*tr_j[0],tr_p*tr_j[1],tr_p*tr_j[2]]
    if log:
        return np.log(tr/np.sum(tr))
    if not log:
        return tr/np.sum(tr)


def transition_matrix_2d_5(leftright1,leftright2,sep2,sep3,log=True):
    '''
    generate a symmetric transition matrix for 0 memory and 1 detector, allowing 2 jumps either side
    args
    -------
    par1: float
        ratio of left and rightjump elements to the centre jump element of the transition matrix
    par2: float
        ratio detector 1 being in left or right bin compared to reference
    par3: float
        ratio detector 2 being in left or right bin compared to reference
    kwargs
    ------
    log: bool
        return the log of the transition matrix if True, just probabilities if false
    returns 
    -------
    tr: array
        3x3 array of transition probabilities 
    '''
    j1 = 1.
    jlr1 = j1/leftright1
    jlr2 = j1/leftright2
    plr1 = j1/sep2
    plr2 = j1/sep3
    tr_j= np.array([jlr2,jlr1,j1,jlr1,jlr2])
    tr_p1 = np.array([plr1,plr1,j1,plr1,plr1])
    tr_p2 = np.array([plr2,plr2,j1,plr2,plr2])
    tr_p = np.array([tr_p1[0]*tr_p2,tr_p1[1]*tr_p2,tr_p1[2]*tr_p2,tr_p1[3]*tr_p2,tr_p1[4]*tr_p2])
    tr = [tr_p*tr_j[0],tr_p*tr_j[1],tr_p*tr_j[2],tr_p*tr_j[3],tr_p*tr_j[4]]
    if log:
        return np.log(tr/np.sum(tr))
    if not log:
        return tr/np.sum(tr)


def track_power(track,data):
    '''
    returns an array the power along any given track 
    args
    -------
    track: array
        the track found by viterbi
    data: 2d array
        data to search through
    returns
    -------
    val: array
        power along given track
    '''
    val = []
    for i,j in enumerate(track):
        val.append(data[i,j])
    return val

def stat_on_path(track,data):
    '''
    returns the viterbi statistic along a given path (currently assumes flat transition matrix) 
    args
    -------
    track: array
        the track found by viterbi
    data: 2d array
        data to search through
    returns
    -------
    val: array
        stat along track
    '''
    # get summed stat along trck
    val = np.array([data[i,j] for i,j in enumerate(track)])
    
    # calculate difference between this and element before
    norm_val = np.insert(np.diff(val), 0, val[0])
    del val
    return norm_val


def ft_power(track):
    f0 = 2./86164.09
    fs = 1./1800
    w = np.exp(-2*np.pi*1j*f0*np.arange(len(track))/fs)
    power = np.abs(np.sum(track*w))**2

def fft_stat(data,paths):
    """ Statistic taking the fft of the track at 1 siderial day"""
    path = [i for i in paths for j in range(48)]
    power = []
    for i,j in enumerate(path[:len(data)]):
        power.append(data[i][j])
    f0 = 1./86164.09
    f1 = 1./85000
    f2 = 1./89000
    signal = np.array(power)
    fs = 1./1800
    w = np.exp(-2*np.pi*1j*f0*np.arange(len(signal))/fs)
    w_back1 = np.exp(-2*np.pi*1j*f1*np.arange(len(signal))/fs)
    w_back2 = np.exp(-2*np.pi*1j*f2*np.arange(len(signal))/fs)
    power_stat = abs(sum(signal*w))**2
    power_back1 = abs(sum(signal*w_back1))**2
    power_back2 = abs(sum(signal*w_back2))**2
    power_back = (power_back1+power_back2)/2.0
    return power_stat/power_back

def highpass_filter(data, cut, fs, order=5):
    """
    simple highpass filter the data with given cutoff and samplin frequency
    """
    nyq = 0.5*fs
    cut1 = cut / nyq
    b, a = butter(order, cut1, btype='highpass')
    y = filtfilt(b, a, data)
    return y

def res(track,path):
    """
    find the summed residual between the viterbi track and the pulsars path
    """
    separation = 0
    for i in range(len(track)):
        separation += np.abs(track[i]-path[i])
    return separation

def frac_off(track,path,noise = None,gap_val = 2):
    """
    find the fraction of track elements which are not in the same frequency bin as the injected signal
    """
    frac = 0
    track_len = 0
    if noise is None:
        pass
    else:
        for i in range(len(noise)):
            if np.isnan(noise[i]) == False or noise[i] == gap_val:
                track_len += 1
            else:
                pass
        
    for i in range(len(track)):
        if noise is not None:
            if track[i] == path[i]:
                if np.isnan(noise[i]) == False or noise[i] == gap_val:
                    frac += 1
                else:
                    pass
            else:
                pass
        else:
            if track[i] == path[i]:
                frac += 1
            else:
                pass
    if noise is None:
        return frac/float(len(track))
    else:
        return frac/float(track_len)
    

def find_rmeds(pul_track,vit_track):
    """
    find the rms (root median squared)of the difference of two tracks
    Parameters
    ------------
    pul_tracks: array
         array of signal tracks, can be single array or list of arrays
    vit_tracks: array
         array of tracks found by search, can be single array of list of arrays
    Returns
    ---------------
    diff: float
        value of rms for list of input tracks
    """
    diff = []
    for elem in range(len(pul_track)):
        pathsqs = ((np.array(pul_track[elem])-np.array(vit_track[elem]))**2)
        diff.append(np.sum(np.median(np.array(pathsqs))))
        #diff.append(1./len(pathsqs)*np.sum(np.array(pathsqs)))

    return np.sqrt(1./(len(diff))*np.sum(diff))

def find_rms(pul_track,vit_track):
    """
    find the rms (root median squared)of the difference of two tracks
    Parameters
    ------------
    pul_tracks: array
         array of signal tracks, can be single array or list of arrays
    vit_tracks: array
         array of tracks found by search, can be single array of list of arrays
    Returns
    ---------------
    diff: float
        value of rms for list of input tracks
    """
    diff = []
    for elem in range(len(pul_track)):
        pathsqs = ((np.array(pul_track[elem])-np.array(vit_track[elem]))**2)
        #diff.append(np.sum(np.median(np.array(pathsqs))))
        diff.append(1./len(pathsqs)*np.sum(np.array(pathsqs)))

    return np.sqrt(1./(len(diff))*np.sum(diff))





import lalpulsar
from .sft import SFT
import lal
import numpy as np
import time

class LoadSFT:
    

    def __init__(self,sftpath,fmin=None,fmax=None,norm=False,summed=False,filled=False,remove_sft=True,save_rngmed=False,tmin=None,tmax=None,vetolist = None, verbose=False):
        """
        Load an SFT from multiple detectors
        args
        -----------------
        sftpath: str or np.ndarray
            path to the sft files, for multiple files separate by semicolon, 'filename1;filename2;.....' can input from multiple detectors
            or input an np.ndarray of shape (Ndetectors, Ntimesteps, Nfreqbins)
        fmin: float
            minimum frequency to load from sft
        fmax: float
            maximum frequency to load from sft
        norm: bool or int
            normalise sft to running median, if integer running median of that many bins
        summed: bool or int
            sum normalised spectrograms over number of bins (if True default 48 (1day of 1800s)) or set to value
        filled: bool
            fill the gaps in the sfts with the expected value (2 for normalised sfts, nan for sfts)
        remove_sft: bool
            remove original sft after normalising to running median
        save_rndmed: bool
            save the running median as an estimate of noise floor
        tmin: float
            start time for sfts
        tmax: float
            end time for sfts
        vetolist: list
            list of frequency bins to set to expected value (2 or nan) (and +/- 3 bins of selected bin)
        det_names: list
            list of detector names (only necessary when using np.ndarray as input to sftpath)
        """
        self.verbose = verbose

        if type(sftpath) is str:
            # load the sft
            self.get_sft(sftpath,fmin=fmin,fmax=fmax,tmin=tmin,tmax=tmax)
        elif type(sftpath) is dict:
            self.get_sft_from_array(sftpath, fmin=fmin, fmax=fmax, tmin=tmin, tmax=tmax)
        else:
            raise Exception(f"Please use a type str or np.ndarray for the input of sftpath, you used type {type(sftpath)}")

        # normalise sft to the running median
        if norm is not None:
            if norm == True:
                # set number of bins to running median to 100 as default
                norm = 100
                self.norm_rngmed(med_width=norm, remove_sft = remove_sft,save_rngmed=save_rngmed)
            elif norm == False:
                pass
            else:
                # set the number of bins for running median
                self.norm_rngmed(med_width=norm, remove_sft = remove_sft,save_rngmed=save_rngmed) 

        # set the frequency bins to veto, i.e. set to mean of spectrogram
        if vetolist is not None:
            self.veto_freq_bins(vetolist)

        # fill the gaps in the sfts
        if filled:
            # fills gaps with mean of chi2(2) distribution
            self.fill_gaps()
            # fills gaps at start and end of each detector to align sfts
            self.align_sfts()

        # sum sfts over given range
        if type(summed) is bool:
            if summed:
                self.sum_sfts()
                self.n_summed_sfts = 48
            else:
                self.n_summed_sfts = 1
        elif type(summed) is int:
            self.sum_sfts(nsfts=summed)
            self.n_summed_sfts = summed
        else:
            print(f"WARNING: Summed must be of type int, you have {type(summed)}")


    def get_sft(self,sftpath,fmin=None,fmax=None,tmin=None,tmax=None,vetolist = None):
        '''
        load an sft to a numpy array, 
        args
        -------
        sftpath : string
            path to the sft file
        fmin    : float
            minimum frequency
        fmax    : float
            maximum frequency
        tmin    : float
            min time
        tmax    : float
            max time
        '''
        if self.verbose:
            start_time = time.time()
        # set up contraints for data
        constraints = lalpulsar.SFTConstraints()

        # set fmin and fmax to all sfts as default
        if fmin is None:
            fmin = -1
        if fmax is None:
            fmax = -1

        # only select sfts with start time (epoch) in range tmin-tmax
        if tmin is not None:
            self.tmin_gps = lal.LIGOTimeGPS(int(tmin),0)
            constraints.minStartTime=self.tmin_gps
        if tmax is not None:
            self.tmax_gps = lal.LIGOTimeGPS(int(tmax),0)
            constraints.maxStartTime=self.tmax_gps

        if not type(sftpath) == str:
            sftpath = ";".join([s.strip() for s in sftpath])
        # create cataologue of sfts under specified constraints
        catalogue = lalpulsar.SFTdataFind(sftpath,constraints)

        # load the sfts from the catalogue above (currently uses multi SFTs as there is a memory leak in LoadSFTs)
        # sfts = lalpulsar.LoadSFTs(catalogue,fmin,fmax)
        sfts = lalpulsar.LoadMultiSFTs(catalogue,fmin,fmax)
        # keep sft set
        #self.sfts_old = sfts

        # define length of sfts in sft index and number of frequency bins
        # N = sfts.length
        self.det_names = []
        tsft = []
        nbins = []
        for det in sfts.data:
            tsft.append(1.0/det.data[0].deltaF)
            nbins.append(det.data[0].data.length)
        
        if len(set(tsft)) > 1:
            print("Warning: tsft not the same between detectors.")
            
        if len(set(nbins)) > 1:
            print("Warning: different detectors do not have the same number of frequency bins")
        
        for det in sfts.data:
            # get detectors name
            detname = det.data[0].name
            self.det_names.append(detname)

            # initialise SFT for detector
            data = SFT()

            # set parameters of sft
            data.nsft = det.length
            data.nbins = det.data[0].data.length
            data.delta_f = det.data[0].deltaF
            data.f0 = det.data[0].f0
            data.tsft = 1.0/det.data[0].deltaF

            #define range of frequency bin centers
            data.frequencies = np.arange(data.nbins)*data.delta_f + data.f0
            
            # create empty arrays for likleyhoods and epochs
            data.sft = np.zeros((data.nsft,data.nbins)).astype(np.complex_)
            data.epochs = np.zeros(data.nsft)
            # set fmin and fmax
            data.fmin = det.data[0].f0
            data.fmax = data.fmin + data.nbins/data.tsft

            #for i,sft in enumerate(sfts.data):
            for i,sft in enumerate(det.data):
                # fill sft with data
                data.sft[i,:] = sft.data.data
                # record epoch of each sft
                data.epochs[i] = sft.epoch
                
            # save sft for detector
            setattr(self,detname,data)
        if self.verbose:
            print("Load SFTs time: ", time.time() - start_time)

    def get_sft_from_array(self,sftpath,fmin=None,fmax=None,tmin=None,tmax=None,vetolist = None):
        '''
        load an sft to a numpy array, 
        args
        -------
        sftpath :dict
            dict containing detector dicts
        detector: string
            which detector i.e 'H1'
        fmin    : float
            minimum frequency
        fmax    : float
            maximum frequency
        tmin    : float
            min time
        tmax    : float
            max time
        '''
        if self.verbose:
            start_time = time.time()
        temp_nsft = []
        temp_nbin = []
        self.det_names = []

        for name,sdict in sftpath.items():
            # get detectors name
            detname = name
            self.det_names.append(detname)

            assert len(sdict["sft"]) == len(sdict["epochs"]), f'Timestamps (Epochs) should match the length of the SFT, sftlen:{len(sdict["sft"])}, epochlen:{len(sdict["epochs"])}'

            # initialise SFT for detector
            data = SFT()

            sftshape = np.shape(sdict["sft"])
            # set parameters of sft
            data.nsft = sftshape[0]

            #define range of frequency bin centers
            if "frequencies" in sdict:
                data.frequencies = sdict["frequencies"]
                data.nbins = len(sdict["frequencies"])
                data.f0 = sdict["frequencies"][0]
                data.delta_f = sdict["frequencies"][1] - sdict["frequencies"][0]
                data.tsft = 1.0/data.delta_f
            elif "f0" in sdict and "delta_f" in sdict:
                data.nbins = sftshape[1]
                data.delta_f = sdict["delta_f"]
                data.f0 = sdict["f0"]
                data.tsft = 1.0/sdict["delta_f"]
                data.frequencies = np.arange(data.nbins)*data.delta_f + data.f0
            else:
                raise Exception(f"Please define the frequencies or f0 and delta_f in the input dict, currently only {sdict.keys()}")
            
            temp_nbin.append(data.nbins)
            temp_nsft.append(data.nsft)

            if len(set(temp_nsft)) > 1:
                print("Warning!: tsft not the same between detectors.")
            
            if len(set(temp_nbin)) > 1:
                print("Warning!: different detectors do not have the same number of frequency bins")

            # create empty arrays for likleyhoods and epochs
            data.sft = np.zeros((data.nsft,data.nbins)).astype(np.complex_)
            data.epochs = np.zeros(data.nsft)
            # set fmin and fmax
            data.fmin = data.f0
            data.fmax = data.f0 + data.nbins/data.tsft


            data.sft = sdict["sft"].astype(np.complex_)
            data.epochs = sdict["epochs"]
                
            # save sft for detector
            setattr(self,detname,data)
        if self.verbose:
            print("Load SFTs time: ", time.time() - start_time)
            
   
    def norm_rngmed(self,med_width=100,remove_sft = False, save_rngmed=False):
        """
        normalise the SFT to the running median and save the output power
        args
        ---------------
        med_width: int (optional)
            number of bins to use for running median, default 100
        remove_sft: bool
            remove the original complex sfts save only the normalised power, default false
        save_rngmed: bool
            save the running median to use as noise floor later, default false
        """
        if self.verbose:
            start_time = time.time()

        for det in self.det_names:
            # get sft for detector
            sfts = getattr(self,det)
            
            # create empty vectors for sft and normalised power
            sig2 = lal.CreateREAL8FrequencySeries(None,0,0,1./sfts.tsft,lal.SecondUnit,sfts.nbins)
            c_sft = lal.CreateCOMPLEX8FrequencySeries(None,0,0,1./sfts.tsft,lal.SecondUnit,sfts.nbins)
            sfts.norm_sft_power = np.zeros((sfts.nsft,sfts.nbins))

            if save_rngmed:
                sfts.rng_med = np.zeros((sfts.nsft,sfts.nbins))

            for i,sft in enumerate(sfts.sft):
                # fill sig2 structure with running medians 
                c_sft.data.data = sft.astype('complex64')
                lalpulsar.SFTtoRngmed(sig2,c_sft,med_width)
                if save_rngmed:
                    sfts.rng_med[i,:] = sig2.data.data
                # fill likleyhood arrays with sft data normalised to running median (multiplied by two to be chi2 distribution)
                sfts.norm_sft_power[i,:] = 2*np.abs(c_sft.data.data/np.sqrt(sig2.data.data))**2

            if remove_sft:
                del sfts.sft

        if self.verbose:
            print("Normalise SFTs time: ", time.time() - start_time)
                
                
    def veto_freq_bins(self,vetolist):
        """
        If frequency bins to be vetoed, set to mean of nornalised chi2 for spectrogram power or nans for sfts
        Vetoes the input bin and +/- 2 bins 
        args
        -----------
        vetolist: list
            list of frequency bins to be vetoes
        """
        for det in self.det_names:
            sft = getattr(self,det)

            if vetolist is not None:
                if len(np.shape(vetolist)) == 0:
                    vetolist = [vetolist]
                for k in list(vetolist):
                    j = int(np.round((k-sft.fmin)*sft.tsft))
                    # set surrounding bins to veto
                    for i in [j-2,j-1,j,j+1,j+2]:
                        try:
                            if hasattr(sft,"sft"):
                                sft.sft[:,i] = np.ones(sft.nsft)*np.nan
                            if hasattr(sft,"norm_sft_power"):
                                sft.norm_sft_power[:,i] = np.ones(sft.nsft)*2
                            if hasattr(sft,"summed_norm_sft_power"):
                                sft.summed_norm_sft_power[:,i] = np.ones(sft.nsft)*96
                        except IndexError: 
                            pass

            
    def fill_gaps(self):
        '''
        fill data gaps with twos, it is filled with twos as it should have a mean of two, this becomes useful when summing over a day.
        args
        ------
        epochs     : array
            start times of current sfts
        likleihootd : array
            sft likelihood arrays
        tsft       : float
            length of sfts
        returns
        ---------
        new_index : new likleyhood array with time gaps filled with twos
        sft_index : index of each sft
        new_epoch : epochs after data is filled
        '''
        
        if self.verbose:
            start_time = time.time()

        for det in self.det_names:
            sft = getattr(self,det)
            # set epochs index for each sft
            sft_index = np.floor((sft.epochs - sft.epochs[0])/sft.tsft).astype('int')

            if hasattr(sft,"sft"):
                if sft.sft is not None:
                    # set equally spaced epochs from start time to end time
                    new_sft = np.array(np.ones((sft_index[-1]+1,sft.sft.shape[1]))*2).astype("complex")
                    # fill appropriate epochs with real data
                    for i,idx in enumerate(sft_index):
                        new_sft[idx,:] = sft.sft[i,:]
                    sft.sft = new_sft

            if hasattr(sft,"norm_sft_power"):
                if sft.norm_sft_power is not None:
                    new_sft = np.ones((sft_index[-1]+1,sft.norm_sft_power.shape[1]))*2
                    for i,idx in enumerate(sft_index):
                        new_sft[idx,:] = sft.norm_sft_power[i,:]
                    sft.norm_sft_power = new_sft

            if hasattr(sft,"rng_med"):
                if sft.rng_med is not None:
                    new_sft = np.ones((sft_index[-1]+1,sft.rng_med.shape[1]))*np.nan
                    for i,idx in enumerate(sft_index):
                        new_sft[idx,:] = sft.rng_med[i,:]
                    sft.rng_med = new_sft

            # rerun summing of sfts
            if hasattr(sft,"summed_norm_sft_power"):
                sft.sum_sft(nsfts=self.n_summed_sfts)

            # define new epochs for filled data
            new_epoch = np.arange(len(new_sft))*sft.tsft + sft.epochs[0]
            sft.epochs = new_epoch
            sft.nsft = len(sft.epochs)
        if self.verbose:
            print("Fill gaps time: ", time.time() - start_time)


    def align_sfts(self):
        """
        fill the start and end portions of sfts such that each detector has the same length of data
        """
        if self.verbose:
            start_time = time.time()

        # define minimim and maximum epochs
        min_ep = np.inf
        max_ep = -np.inf
        min_det = None
        max_det = None
        tsft = None
        for det in self.det_names:
            det_data = getattr(self,det)
            tsft = det_data.tsft
            min_ep_temp = min(det_data.epochs)
            max_ep_temp = max(det_data.epochs)
            if min_ep_temp < min_ep:
                min_ep = min_ep_temp
                min_det = det
            if max_ep_temp > max_ep:
                max_ep = max_ep_temp
                max_det = det

        # define total number of epochs from minmax and tsft
        num_ep = np.floor((max_ep - min_ep)/tsft).astype("int")

        for det in self.det_names:
            sft = getattr(self,det)
            # find epoch indicies
            sft_index = np.floor((sft.epochs - min_ep)/sft.tsft).astype('int')

            if hasattr(sft,"sft"):
                if sft.sft is not None:
                    new_sft = np.array(np.ones((num_ep+1,sft.sft.shape[1]))*2).astype("complex")
                    for i,idx in enumerate(sft_index):
                        new_sft[idx,:] = sft.sft[i,:]
                    sft.sft = new_sft

            if hasattr(sft,"norm_sft_power"):
                if sft.norm_sft_power is not None:
                    new_sft = np.ones((num_ep+1,sft.norm_sft_power.shape[1]))*2
                    for i,idx in enumerate(sft_index):
                        new_sft[idx,:] = sft.norm_sft_power[i,:]
                    sft.norm_sft_power = new_sft

            if hasattr(sft,"rng_med"):
                if sft.rng_med is not None:
                    new_rng = np.ones((num_ep+1,sft.rng_med.shape[1]))*2
                    for i,idx in enumerate(sft_index):
                        new_rng[idx,:] = sft.rng_med[i,:]
                    sft.rng_med = new_rng


            new_epoch = np.arange(len(new_sft))*sft.tsft + sft.epochs[0]
            sft.epochs = new_epoch
            sft.nsft = len(sft.epochs)

            if hasattr(sft,"summed_norm_sft_power"):
                sft.sum_sft(nsfts=self.n_summed_sfts)

        if self.verbose:
            print("Align SFTs time: ", time.time() - start_time)
        
            
    def sum_sfts(self,sum_type="norm_sft_power",gap_val = 2,nsfts=48):
        '''
        takes input sfts with tsft=1800s and returns sum of 48 sfts, i.e summing over a day
        and returns fraction of real data in each sum.
        args
        -------
        gap_val: double
            value used when there are gaps in the data (for calculating fraction)
        returns
        -----------
        data_av: array
            data summed over every day
        fraction: array
            fraction of each day which contained real data
        '''
        if self.verbose:
            start_time = time.time()

        for dt in self.det_names:
            data = getattr(self,dt)
            data.sum_sft(sum_type=sum_type,gap_val=gap_val,nsfts=nsfts,remove_original=False)

        if self.verbose:
            print("Sum SFTs time: ", time.time() - start_time)

    def downsamp_frequency(self,data_type="summed_norm_sft_power",stride = 2,remove_original=False):
        '''
        downsample frequency of sfts by taking mean of a number (stride) of frequency bins starting from base of band
        args
        -------
        stride: int
            number of frequency bins to take mean of and downsample by
        '''
        if self.verbose:
            start_time = time.time()

        for dt in self.det_names:
            data = getattr(self,dt)
            data.downsamp_frequency(data_type=data_type,stride=stride,remove_original=remove_original)

        if self.verbose:
            print("Downsamp frequency time: ", time.time() - start_time)




from __future__ import absolute_import
import numpy as np
import timeit
import lal
import lalpulsar
from lalpulsar import simulateCW
from scipy.special import fresnel
import scipy.stats as st
import time
from astropy.time import Time
import os
import copy
import subprocess
from .sft import SFT
from .timeseries import TimeSeries
from .tools import download_ephemeris_file, LAL_EPHEMERIS_URL


class GenerateSignal:
    
    def __init__(self,alpha=None,delta=None,psi=None,phi0=None,cosi=None,h0=None,f=None,tref=None,earth_ephem=None,sun_ephem=None,snr=None):
        """
        Set up parameters to generate a CW signal in noise
        args
        ---------
        alpha: float
            right ascencion for object, [0, 2pi]
        delta: float
            declination of object [-pi/2, pi/2]
        psi: float
            polarisation of object [0,2pi]
        phi0: float
            initial phase of signal
        cosi: float
            cosine of the inclination of star [0,1]
        h0: float
            Amplitude of the gravitational wave
        f: list or array
            2 component list of frequency parameters [f, fdot]
        tref: float
            reference time for phase (default tstart)
        earth_ephem: string
            path to earth ephemeris file
        sun_ephem: string
            path to sun ephemeris file
        snr: float
            SNR of signal, overwrites the h0 parameter
        """
        if earth_ephem is None:
            self.earth_ephem = "earth00-40-DE430.dat.gz"
        else:
            self.earth_ephem = earth_ephem
        if sun_ephem is None:
            self.sun_ephem = "sun00-40-DE430.dat.gz"
        else:
            self.sun_ephem = sun_ephem
        
        self.alpha = alpha
        self.delta = delta
        self.psi   = psi
        self.phi0  = phi0
        self.cosi  = cosi
        self.h0    = h0
        self.f     = f
        self.tref = None
        
        self.gaps = False
        self.snr = snr
        self.edat = None
        self.antenna_pattern = True
        self.baryinputs = {}
        self.siteinfo = {}
        self.earth = lalpulsar.EarthState()
        self.emit = lalpulsar.EmissionTime()
        
    def param_check(self):
        
        for i,v in enumerate(params[:-2]):
            if np.isnan(v):
                raise Exception("parameter {} is not a valid value".format(params_names[i]))
                
        if self.tref is None:
            print("No reference time: Pulsar referece time same as sft start time")
            self.tref = self.tstart

        
    def get_baryinput(self, detector):
        """
        Load in the baryinputs prior to running
        """
        baryinput = lalpulsar.BarycenterInput()
        if detector == 'SSB':
            # set up the for reference frame at center of earth                                                                              
            baryinput.site.location[0] = 0./lal.C_SI
            baryinput.site.location[1] = 0./lal.C_SI
            baryinput.site.location[2] = 0./lal.C_SI
        else:
            # get the information on specified detector                                                                                      
            # set up the barycenter for given detector    
            siteinfo = lalpulsar.GetSiteInfo(detector)   
            self.siteinfo[detector] = siteinfo                                                                             
            baryinput.site = siteinfo
            baryinput.site.location[0] = baryinput.site.location[0]/lal.C_SI
            baryinput.site.location[1] = baryinput.site.location[1]/lal.C_SI
            baryinput.site.location[2] = baryinput.site.location[2]/lal.C_SI

        return baryinput
        
        
    def detector_velocity(self,edat,epoch,detector):
        '''                                                                                                                                  
        get the three dimensional velocity of the earth                                                                                      
        args                                                                                                                                 
        --------                                                                                                                             
        edat  : the position of the earth and sun at a given time                                                                            
        epoch : the time of the start of the observation                                                                                     
        detector : the detector i.e. 'H1'                                                                                                    
                                                                                                                                         
        returns                                                                                                                              
        --------                                                                                                                             
        emit.vDetector : velocity of detector at arrival time (epoch) in cartesian coords                                                    
        '''

        # initialise the barycenter structure   
        if detector not in self.baryinputs.keys():
             self.baryinputs[detector] = self.get_baryinput(detector)                                                                                             

        # distance to the source in [s^-1]                                                                                                   
        self.baryinputs[detector].dInv = 0.0

        # source right ascension and declination in radians                                                                                  
        self.baryinputs[detector].alpha = 0.0
        self.baryinputs[detector].delta = 0.0

        # finds position and orientation of the earth at arrival time (epoch) for earth state and ephemerus data                             
        lalpulsar.BarycenterEarth(self.earth,epoch,edat)

        # transforms arrival time (epoch) into pulse emission time                                                                           
        lalpulsar.Barycenter(self.emit,self.baryinputs[detector],self.earth)

        return self.emit.vDetector

    def get_detector_velocities(self,epochs: np.array,det: str,edat = None):
        """
        get the velocities of the detector
        args
        -------------
        epochs: array or list
            list of the epoch times you would like to get their velocities
        det: detector
            which detector you would like this data for
        edat: LALEphemeris (optional)
            ephemeris data for solar system (can be auto loaded with get_edat)
        """
        if edat is not None:
            edat = edat
        elif self.edat is None:
            self.get_edat()
            edat = self.edat
        else:
            edat = self.edat

        self.det_vels = np.array(np.zeros((len(epochs),3))*np.nan)
        freqs = np.zeros(len(epochs))
        for i,epoch in enumerate(epochs):
            # get the detector velocity at certain epoch and ephemeris data                                                                  
            #epoch = lal.LIGOTimeGPS(epoch,0)                                                                                                
            vel = self.detector_velocity(edat,epoch,det)
            self.det_vels[i] = vel


    def get_pulsar_path(self,epochs,det,edat=None):
        """
        find the pulsar path in the time frequency plane, given the epochs, parameters and ephemeris data
        args                                                                                                                                 
        ------                                                                                                                               
        epochs : start times of each sft                                                                                                     
        edat   : ephermeris data for earth and sun                                                                                           
        det    : detector i.e 'H1'                                                                                                           
        params : parameters of the pulsar                                                                                                    
        """
        if not hasattr(self, "det_vels") or len(epochs) != len(self.det_vels):
            self.get_detector_velocities(epochs,det,edat)
        
        cosd = np.cos(self.delta)
        cosa = np.cos(self.alpha)
        sind = np.sin(self.delta)
        sina = np.sin(self.alpha)

        vDotn_c = cosd * cosa * self.det_vels[:,0] + cosd * sina  * self.det_vels[:,1] + sind * self.det_vels[:,2]

        if self.tref is None:
            self.tref = epochs[0]

        timeDiff = epochs - self.tref
        fhat = self.f[0] + timeDiff*self.f[1]
        
        return fhat * (1 + vDotn_c)

    def gps_to_gmst_rad(self, epochs):
        """ Taken amd vectorised from LAL code,  https://lscsoft.docs.ligo.org/lalsuite/lal/_x_l_a_l_sidereal_time_8c_source.html#l00062
        !!!!!!!!!! will definitely need updating !!!!!!!!!!!!!!!
        """
        #print("using custom GMST code taken from LAL")
        as_time = Time(epochs, format="gps", scale="utc")
        # keep track of larger part
        t_hi = (as_time.jd - 2451545.0)/36525.0
        # keep track of nanosecond part separately
        t_lo = (as_time.gps % 1e-9)/(1e9 * 36525.0 * 86400.0)
        t = t_hi + t_lo
        
        sidereal_time = (-6.2e-6 * t + 0.093104) * t * t + 67310.54841
        sidereal_time += 8640184.812866 * t_lo
        sidereal_time += 3155760000.0 * t_lo
        sidereal_time += 8640184.812866 * t_hi
        sidereal_time += 3155760000.0 * t_hi
        
        return sidereal_time * np.pi / 43200.0

    def av_antenna(self,alpha,delta,det,tstart,nsft,tsft,antenna=True,use_lal=False,average = True):
        """
        Calculates the time averages of the antenna patterns
        args
        ----------
        alpha: float
            right ascension
        delta: float
            declination
        det: string
            detector, 'H1', 'L1' etc
        T: float
            time of observation in seconds
        returns
        -----------
        A: float
        B: float
        C: float
        """
        # detector parameters in radians, lamb = detectro latitiude, L = longitude,
        # gamma = orientation of detector arms, zeta = angle between arms

        # need to change to get from lalpulsar
        det_params = {"H1": {"lamb": 46.45*np.pi/180., "L": 119.41*np.pi/180., "gamma": 171.8*np.pi/180., "zeta": 90.*np.pi/180.},
                      "L1": {"lamb": 30.56*np.pi/180., "L": 90.77*np.pi/180., "gamma": 243.0*np.pi/180., "zeta": 90*np.pi/180.},
                      "V1": {"lamb": 43.63*np.pi/180., "L": -10.5*np.pi/180., "gamma": 116.5*np.pi/180., "zeta": 90*np.pi/180.}}

        # this is not the source polarisation, that will be used later
        psi = 0.0

        if not antenna:
            return np.array([[1,1,1]])
        else:
            if det not in self.siteinfo.keys():
                self.baryinputs[det] = self.get_baryinput(det)  
            #detr = lalpulsar.GetSiteInfo(det)
            epochs = np.arange(tstart,tstart+tsft*nsft,tsft) + 0.5*tsft
            #gmst_rad = [lal.GreenwichMeanSiderealTime(i) for i in epochs]
            gmst_rad = self.gps_to_gmst_rad(epochs)
            am = [lal.ComputeDetAMResponse(self.siteinfo[det].response, alpha, delta, 0.0, i) for i in gmst_rad]
            fraczeta = (1./np.sin(det_params[det]["zeta"]))
            cos2psi = np.cos(2*psi)
            sin2psi = np.sin(2*psi)
            #A,B,C = 0,0,0
            ABC = []
            for i in range(len(epochs)):
                av,bv = 0,0
                fpl,fcr = am[i]
                av = fraczeta*(fpl*cos2psi - fcr*sin2psi)
                bv = fraczeta*(fcr*cos2psi + fpl*sin2psi)
                ABC.append((2*av*av,2*bv*bv,2*av*bv))
                del av,bv

            if average:
                return np.sum(ABC,axis=0) * 1./len(epochs)
            else:
                return np.array(ABC)

    def get_snr2(self,epochs=None,alpha=None, delta=None, psi=None, phi0=None, cosi=None, Sn=None, det=None, tstart=None, nsft=None, tsft=None, h0 = None,snr = None,antenna=True):
        """
        Calculates the snr or h0, if h0 input calculates snr if snr input calculates h0
        if snr<0 or None will find snr, else will find h0
        args
        ----------
        alpha: float
            right ascension
        delta: float
            declination
        psi: float
            polarisation of the wave
        phi0: float
            phase of the wave
        cosi: float
            cos of the inclination angle iota
        Sh: float or array
            noise spectral density at frequency of signal, not sqrt(Sh), or array if Sh differes with time
            if Sh == None, will calculate Sh from snr and h0
        det: string
            detector, 'H1', 'L1' etc
        tstart: float
            start time of observation
        T: float
            time of observation in seconds
        h0: float
            amplitude of the wave
        snr: float
            signal to noise 
        antenna: bool
            true is antenna pattern

        returns
        -----------
        snr_2 or h0: float
            The signal to noise squared or h0

        """
        params = {'alpha':alpha, 'delta':delta, 'psi':psi,'phi0':phi0,'cosi':cosi,'tstart':tstart,'nsft':nsft,'tsft':tsft,'h0':h0}

        for i,j in params.items():
            if j is None:
                if i in ["tstart","nsft"]:
                    pass
                else:
                    params[i] = getattr(self, i)
            else:
                pass

        ap = (1./2)*(1+params["cosi"]*params["cosi"])
        ac = params["cosi"]
        
        cos2phi0 = np.cos(2*params["phi0"])
        sin2phi0 = np.sin(2*params["phi0"])
        sin2psi = np.sin(2*params["psi"])
        cos2psi = np.cos(2*params["psi"])

        A1 = ap*cos2phi0*cos2psi - ac*sin2phi0*sin2psi
        A2 = ap*cos2phi0*sin2psi + ac*sin2phi0*cos2psi
        A3 = -ap*sin2phi0*cos2psi - ac*cos2phi0*sin2psi
        A4 = -ap*sin2phi0*sin2psi + ac*cos2phi0*cos2psi

        alp1 = A1*A1+A3*A3
        alp2 = A2*A2+A4*A4
        alp3 = A1*A2+A3*A4

        if epochs is None:
            if tstart is None or tsft is None or nsft is None:
                raise Exception("Please define epochs or (tstart,nsft,tsft)")
            else:
                epochs = np.linspace(params["tstart"], params["tstart"] + params["nsft"]*params["tsft"], params["nsft"])
        if params["tsft"] is None:
            params["tsft"] = epochs[1] - epochs[0]
        if params["tstart"] is None:
            params["tstart"] = epochs[0]
        if params["nsft"] is None:
            params["nsft"] = len(epochs)

        if Sn is None:
            raise Exception("Please define the noise floor Sn")
            
        if len(np.shape(Sn)) == 0:
            Sn = np.ones(len(epochs))*Sn

        if len(epochs) != len(Sn):
            raise Exception("Sn should be same length as epochs")

        ABC = self.av_antenna(alpha=params["alpha"],delta=params["delta"],det=det,tstart=params["tstart"],nsft=params["nsft"],tsft=params["tsft"],antenna=antenna,average=False)
        # nan to num accounts for areas when Sn is 0 or nan and sets the SNR to 0 when no data
        A,B,C = ABC[:,0],ABC[:,1],ABC[:,2]
        snr_2 = np.nan_to_num(((params["h0"]*params["h0"]*1*params["tsft"])/(2.0*Sn))*(alp1*A + alp2*B + 2*alp3*C))
        return snr_2
        
       

    def get_edat(self):
        """
        Get the ephemeris data from supplied ephemeris files, if not defined will download filenames
        """
        try:
            self.edat_p = [self.sun_ephem,self.earth_ephem]
            self.edat = lalpulsar.InitBarycenter(earthEphemerisFile=self.earth_ephem,sunEphemerisFile=self.sun_ephem)
        except Exception as e:
            print("Could not load ephemeris file: {} {}, {}".format(self.earth_ephem, self.sun_ephem, e))
            try:
                self.earth_ephem = download_ephemeris_file(LAL_EPHEMERIS_URL.format(self.earth_ephem))
                self.sun_ephem = download_ephemeris_file(LAL_EPHEMERIS_URL.format(self.sun_ephem))
                self.edat_p = [self.sun_ephem,self.earth_ephem]
                self.edat = lalpulsar.InitBarycenter(earthEphemerisFile=self.earth_ephem,sunEphemerisFile=self.sun_ephem)
            except Exception as e:
                raise IOError("Could not read in ephemeris files: {}".format(e))
        


        
    def get_spectrogram(self,fmin=None,fmax=None,tsft=None,epochs=None,Sn=None,tstart=None,nsft=None,dets=None,snr=None,doppler=True,pulsar_path=None,tref=None,antenna=True,noise_spect=None):
        """
        generates a spectrogram from the signal given some inputs
        
        args
        -----
        fmin: float
            lower frequency bound for spectrogram
        fmax: float
            upper frequency bound for spectrogram
        tsft: float
            length of each sft in seconds
        epochs: array (optional)
            array of time epochs to calculate sfts for, if not defined then these are found from tstart, T and tsft
        Sn: float, array, dict
            noise floors for each detector, if float then each detector has the same noise floor for whole run, can put in same float for each detector as array or and array for each detector as noise floor vaires. Best to put in dictionary saying which detector has which noise floor
        tstart: int (optional)
            start time of observation in seconds, in None this is found from epochs
        nsft: int (optional)
            number of sfts, if None this is found from T and tsft or epochs
        T: float (optional)
            length of data in seconds, nsft*tsft
        dets: str or list (optional)
            list of detectors to generate spectrogram for, if None, then uses keys defined in Sn
        snr: float (optional)
            total SNR of injections, this is the integrated recovered SNR for all detectors
        doppler: bool (optional)
            turn of doppler effects if False, default True
        pulsar_path: array (optional)
            path in frequency that a signal track should take, should be arrays for each detector, default find path from signal parmaeters
        tref: int (optional)
            reference time for antenna pattern, default time is tstart.
        antenna: bool (optional)
            turn off the antenna pattern modulations if False, default True
        noise_spect: dict
            real data to inject a signal into, format {"H1": H1_data, "L1": L1_data}
        
        returns
        ----------
        spt: SimulateSpectrogram class
            simulate spectrogram example attributes if one detector at H1
            H1: dict
                {"spect":spectrogram,"pulsar_path":pulsars track etc}
        
        """
        
        spt = SimulateSpectrogram(self)
        
        spt.__gen_spect__(epochs=epochs,tstart = tstart, nsft=nsft,tsft=tsft,fmin=fmin,fmax=fmax,dets=dets,snr=snr,Sn=Sn,doppler=doppler,antenna=antenna,tref=tref,pulsar_path=pulsar_path,noise_spect=noise_spect)
            
        return spt
    
    def get_timeseries(self,duration = None,tstart=None,tref=None,Sn=None,detectors=None,sample_frequency=4096):
        """
        gets the timeseries using lalpulsar.simulateCW.CWSimulator
        
        """

        timeseries = SimulateTimeseries(self)

        timeseries.gen_timeseries(duration=duration, tstart=tstart, tref=tref, Sn=Sn, detectors=detectors, sample_frequency=sample_frequency)

        return timeseries

class SimulateTimeseries:

    def __init__(self, parent):
        self._parent = parent
        self._parent.get_edat()

    def __getattr__(self, name):
        if name in self._parent.__dict__:
            try:
                return getattr(self._parent, name)
            except AttributeError:
                pass

        if name not in self.__dict__:
            raise AttributeError(name)
        return self.__dict__[name]

    def waveform(self,h0, cosi, freq, f1dot):
        def wf(dt):
            dphi = lal.TWOPI * (freq * dt + f1dot * 0.5 * dt**2)
            ap = h0 * (1.0 + cosi**2) / 2.0
            ax = h0 * cosi
            return dphi, ap, ax
        return wf

    
    def gen_timeseries(self,duration=1800, tstart=None, tref=None, Sn=None, detectors=["H1"], sample_frequency=4096):
                
        if self.h0 is None and self.snr is None:
            raise Exception("Please define either SNR or h0")

        if detectors is None and Sn is None:
            raise Exception("Please define the detectors of noise floor Sn")

        if tref is None:
            tref = self.tstart
        
        if Sn is not None:
            # if noise floor defined use it
            detectors = list(Sn.keys())
            Sn = Sn
            self.Sn = Sn


        if self.snr is not None and self.snr>=0 and self.Sn is not None:
            h0_bar = 1
            snr_bar = 0
            for dt in dets:
                if self.snr != 0:
                    snr_bar += np.nansum(self._parent.get_snr2(epochs=self.epochs,alpha=self.alpha,delta=self.delta,psi=self.psi,phi0=self.phi0,cosi=self.cosi, Sn=self.Sn[dt],det=dt,tsft=self.tsft, h0=h0_bar, antenna=self.antenna_pattern))
            scale = self.snr/np.sqrt(snr_bar)
            self.h0 = h0_bar*scale
        elif self.h0 is not None:
            self.h0 = self.h0


        
        # generate waveform
        wf = self.waveform(self.h0, self.cosi, self.f[0], self.f[1])
        for det in detectors:
            setattr(self,det,TimeSeries())
            timeseries = getattr(self,det)
            
            S = simulateCW.CWSimulator(tref, tstart, duration, wf, 1./sample_frequency, self.phi0, self.psi, self.alpha, self.delta, det, earth_ephem_file=self.earth_ephem,sun_ephem_file=self.sun_ephem)
            
            if Sn is None:
                ltime,signal = S.get_strain(sample_frequency)
            else:
                ltime,signal = S.get_strain(sample_frequency,noise_sqrt_Sh=np.sqrt(Sn[det]))
            
            
            timeseries.timeseries = signal
            timeseries.epochs = ltime
            timeseries.sample_frequency = sample_frequency
            timeseries.delta_t = 1./timeseries.sample_frequency
        return timeseries

class SimulateSpectrogram:
    
    def __init__(self,parent):
        self._parent = parent
        
    def __getattr__(self, name):
        if name in self._parent.__dict__:
            try:
                return getattr(self._parent, name)
            except AttributeError:
                pass

        if name not in self.__dict__:
            raise AttributeError(name)
        return self.__dict__[name]

    def fresnel_power(self,f,f0,tsft,alp):
        '''
        gives the Fresnel integrals for the fourier transform of a signal with changing frequency.
        ---------
        args
        --------
        f: float
            frequency
        f0: float
            central frequency of signal
        tsft: float
            length of an sft in seconds
        alp: float
           rate of change of frequency (\dot{f})
        -------
        returns
        --------
        Sw2: float
            power spectrum of signal at frequency w
        '''
        alp = np.abs(alp)
        sqrtalp = np.sqrt(alp/2.0)*tsft
        frq = 2*(f-f0)/(alp*tsft)
        x1 = sqrtalp*(1+frq)
        x2 = sqrtalp*(1-frq)

        S,C = fresnel([x1,x2])
        S1,S2 = S
        C1,C2 = C
        Sw2 = (1/(2*alp*tsft*tsft))*((C1+C2)*(C1+C2) + (S1+S2)*(S1+S2)) 
        return Sw2

    
    def __gen_spect__(self,fmin=None,fmax=None,tsft=None,epochs=None,Sn=None,tstart=None,nsft=None,dets="H1",snr=None,doppler=False,tref=None,rand=True,antenna=True,pulsar_path=None,noise_spect=None):
        """
        generates a normalised spectrogram, such that the mean of the noise is 2, (i.e. chi2 distribued data) and injectes a signal defined by GenerateSignal.
        -----------
        args
        -----------
        tsft: float
            length of each sft
        fmin: float
            lower frequency band
        fmax: float
            upper frequency band
        tstart: float (optional)
            start time in epoch time (set by first element of epochs if epochs defined)
        T: int (optional)
            length of time in number of sfts (set by epochs and tsft if they are defined)
        tref: int (optional)
            reference time for pulsar parameters, default tstart
        det: string or list
            which detector , 'H1','L1' etc, if multiple detectors ["H1","L1",....] etc
        snr: float (optional)
            snr of signal, if none uses h0 from params to find snr (the recovered SNR should be used as an actual measure of SNR, usually the same as this)
        antenna: bool (optioal)
            True for antenna pattern, false for no antenna pattern (default is True)
        rand: bool (optional)
            use random data or flat background with just signal (default is True)
        doppler: bool (optional)
            turn on or off doppler modulation of signal (default is True)
        Sn: float or array
            either a float for a fixed noise level, or array containing noise floor for each sft where gaps are NaN, if there are multiple (N) detectors Sn should N dimensional, either Nx1 or Nx(nsft) 

        """
        
        #self.param_check()
        # set the epochs and sft parameters from inputs
        self._parent.get_edat()
        self.T = None
        if fmin is None:
            raise Exception("Please set fmin as minimum frequency")
        else:
            self.fmin = fmin
        if fmax is None:
            raise Exception("Please set fmax as maximum frequency")
        else:
            self.fmax = fmax
        if tsft is None:
            raise Exception("Please set an sft length")
        else:
            self.tsft = tsft
    
        if snr is not None:
            self.snr = snr
        else:
            self.snr = None

        # define the noise floor if specified
        if dets is None and Sn is None and noise_spect is None:
            raise Exception("Please define the detectors (dets), noisefloors (Sn) or real data (noise_spect) to simulate")
        elif dets is None and type(Sn) in [dict]:
            dets = list(Sn.keys())
        elif dets is None and noise_spect is not None:
            dets = list(noise_spect.keys())
        elif type(dets) in [str]:
            dets = [dets]

        # if using multiple of the same detector, set the duplicate of noise floors
        if len(set(dets)) == len(dets):
            self.det_names = dets
        elif len(set(dets)) < len(dets):
            sa = set(dets)
            sd = {}
            for i in sa:
                sd.setdefault(i,0)
            self.det_names = []
            for i in dets:
                for j in sd.keys():
                    if i == j:
                        self.det_names.append(j + "_{}".format(sd[j]))
                        sd[j] +=1
                    else:
                        pass
        
            
        if noise_spect is not None:
            self.nsft = len(noise_spect[self.det_names[0]])
        
        # set the epochs for each sft, if it has not been defined already
        if epochs is None:
            if tstart is None or tsft is None and nsft is None:
                raise Exception("Please define either epochs or [tstart,tsft,nsft]")
            self.nsft = nsft
            self.T = nsft*tsft
            self.tstart = tstart
            self.epochs = np.linspace(tstart,tstart+(self.nsft-1)*tsft,self.nsft)
        else:
            self.epochs = epochs
            self.tstart = self.epochs[0]
            self.nsft = len(self.epochs)
            self.tsft = epochs[1] - epochs[0]

        # if the noise floor is not defined then define it, also check that if this is a list then 
        # also set it as an array of length epochs
        # want to define Sn as a dictionary so can be redefined later
        if noise_spect is not None:
            if Sn is not None:
                print("Your value of Sn will not be used, using noise spect instead")
            self.Sn = {}
            # for each detector get the positions where the noise is nan and set this as gap (igner noise floor value at this point as would normalise out anyhow)
            for dt in noise_spect.keys():
                sn_temp = np.ones(len(noise_spect[dt]))
                for ind,val in enumerate(np.mean(noise_spect[dt],axis=1)):
                    if val == 2 or np.isnan(val):
                        sn_temp[ind] = np.nan
                self.Sn[dt] = sn_temp
                del sn_temp
        else:
            # defined noise floor as constant over time is detectors defined but Sn not
            if Sn is None and dets is not None:
                self.Sn = {}
                for det in dets:
                    self.Sn[det] = np.ones(len(self.epochs))
            # if Sn defined check if constant or array
            elif type(Sn) in [dict]:
                if all(name in Sn for name in dets):
                    self.Sn = Sn
                    # for each detector if values is a float, then make this value same length as data otherwise ignore
                    for key in self.Sn.keys():
                        if len(np.shape(self.Sn[key])) == 0:
                            self.Sn[key] = np.ones(len(self.epochs))*self.Sn[key]
                        else:
                            if len(self.Sn[key]) != len(self.epochs):
                                raise Exception("Epochs and noise floor not the same length")
            else:
                raise Exception("Make sure keys for Sn and dets match")

        
        # set the scaled h0 of the signal based of the snr and the noise floor for each epoch and detector
        if self.snr is not None and self.snr>=0 and self.Sn is not None:
            h0_bar = 1
            snr_bar = 0
            for dt in dets:
                if self.snr !=0:
                    snr_bar += np.nansum(self._parent.get_snr2(epochs=self.epochs,alpha=self.alpha,delta=self.delta,psi=self.psi,phi0=self.phi0,cosi=self.cosi, Sn=self.Sn[dt],det=dt,tsft=self.tsft, h0=h0_bar, antenna=self.antenna_pattern))
                #print("SNRBAR1: {}".format(snr_bar))
            scale = self.snr/np.sqrt(snr_bar)
            self.h0 = h0_bar*scale
        elif self.h0 is not None and self.Sn is not None:
            self.h0 = self.h0
        else:
            raise Exeption("Please define either snr or h0 and Sn")

        

        self.harmonic_sum_Sn = 0
        # simulate the spectrogram for each detector
        for dn,dt in zip(self.det_names,dets):
            # set the noise floor if defined
            if noise_spect is not None:
                dt_noise_sp = noise_spect[dt]
            else:
                dt_noise_sp = None
            # generate the spectrogram
            det_data = self.__sim_data__(epochs=self.epochs,tstart=self.tstart,nsft=self.nsft,pulsar_path=pulsar_path,tref=tref,det=dt,antenna=antenna, rand=rand,doppler=doppler,Sn = self.Sn[dt],noise_spect=dt_noise_sp)
            # calcualte the harmonic sum of median noise floors for calulation of depth
            if self.h0 !=0:
                self.harmonic_sum_Sn += det_data.median_Sn 
            setattr(self, '{}'.format(dn), det_data)
            #self.__setattr__("{}".format(dn),det_data)
            del det_data
        # depth of signal as mean of depths from each detector
        if self.h0 == 0:
            self.depth = np.inf
        else:
            self.depth = np.sqrt(1/self.harmonic_sum_Sn)/self.h0 


                

    def __sim_data__(self,pulsar_path=None,tref=None,epochs=None,tstart=None,T=None,nsft=None,det='H1',antenna=True, rand=True,doppler=False,Sn=None,noise_spect=None):
        """
        Simulate data with pulsar signal
        -----------
        args
        -----------
        pulsar_path: array
            indicies of the track the pulsar makes in frequency
        tref: int
            reference time for pulsar parameters, default tstart
        epochs: array
            array of times for each sft
        tstart: float
            start time in epoch time
        T: int
            length of time in number of sfts
        nsft: float
            number of sfts to make
        det: string
            which detector , 'H1','L1' etc
        antenna: bool
            True for antenna pattern, false for no antenna pattern
        rand: bool
            use random data or flat background with just signal
        doppler: bool
            turn on or off doppler modulation of signal
        Sn: float or array
            either a float for a fixed noise level, or array containing noise floor for each sft where gaps are NaN
        noise_spect: 2d array
            set the noise to inject signal 


        ----
        kwargs
        ----
        snr: float
            snr of signal, if none uses h0 from params
        ----------
        returns
        ----------
        dict: 
        {"spect_1","spect_2","freq_track_1","freq_track_2","fmin","fmax","h0","SNR_1","SNR_2","SSB_track_index","depth","freqs","tot_snr"}
        """

        # initialise the SFT
        sft = SFT(tsft=self.tsft)

        #print("input minmaxf", self.fmin, self.fmax)
        # define the parameters of the sft
        #print("multipl", self.fmin*self.tsft, np.round(self.fmin*self.tsft), np.round(self.fmin*self.tsft).astype(int))
        #print("multipl2", self.fmax*self.tsft, np.round(self.fmax*self.tsft), np.round(self.fmax*self.tsft).astype(int))
        self.fmin = np.round(self.fmin*self.tsft).astype(int)/float(self.tsft) # min frequency
        self.fmax = np.round(self.fmax*self.tsft).astype(int)/float(self.tsft) # max frequency
        #print("round minmaxf", self.fmin, self.fmax)

        self.nbins = np.round((self.fmax-self.fmin)*self.tsft).astype(int) # number of frequency bins
        #print("nbins:", self.fmin, self.fmax, self.fmax-self.fmin, (self.fmax-self.fmin)*self.tsft, np.round((self.fmax-self.fmin)*self.tsft))
        self.delta_f = 1./self.tsft # separation of bins in frequency
        self.frequencies = np.arange(self.fmin,self.fmax,sft.delta_f)#[:-1] # bin centers for each bin

        sft.fmin = self.fmin
        sft.fmax = self.fmax
        sft.frequencies = self.frequencies
        sft.nbins = self.nbins
        sft.delta_f = self.delta_f
        sft.nsft = len(epochs)
        sft.epochs = epochs

        # initialise the sft as chi2 with 2 degrees of freedom (or ones if no noise)
        if rand:
            sft.norm_sft_power = st.chi2.rvs(int(2),loc=0,scale=1,size=(sft.nsft,sft.nbins))
        if not rand:
            sft.norm_sft_power = np.ones((sft.nsft,sft.nbins))

        # id the noise if defined from real data, then use this
        if noise_spect is not None:
            sft.norm_sft_power = noise_spect
            nsft = len(noise_spect)

        # set reference time as start of sft
        if not tref:
            self.tref = tstart
        elif tref:
            self.tref = tref

        # define doppler parameters of signal
        params_doppler = lalpulsar.PulsarDopplerParams()
        params_doppler.fkdot[0] = self.f[0]
        params_doppler.fkdot[1] = self.f[1]
        params_doppler.refTime  = self.tref
        params_doppler.Alpha    = self.alpha
        params_doppler.Delta    = self.delta

        # gen the frequency of the signal in the detector and SSB
        if doppler and pulsar_path is None:
            sft.pulsar_path = self._parent.get_pulsar_path(epochs,det)[:nsft]
            pulsar_path_ssb = self._parent.get_pulsar_path(epochs,'SSB')[:nsft]
        elif not doppler and pulsar_path is None:
            sft.pulsar_path = np.ones(nsft)*self.f[0]
            pulsar_path_ssb = None
        else:
            sft.pulsar_path = pulsar_path
            pulsar_path_ssb = None

        # get the bin centers of the signal 
        sft.pulsar_index = np.floor((sft.pulsar_path-self.fmin)*self.tsft) #+ 0.5/self.tsft
        bin_edges = (sft.pulsar_index/self.tsft)+self.fmin


        powers = []
        snr_calc = []

        if self.snr == 0 or self.h0 == 0:
            snr_calc.extend(np.zeros(nsft))
            # if no signal injected then fill gaps with mean of chi2 = 2
            for i,s in enumerate(Sn):
                if np.isnan(Sn[i]):
                    sft.norm_sft_power[i] = np.ones(len(sft.norm_sft_power[i]))*2
        else:
            # gets snr for each epoch, and convets nans where Sn is a gap to snr of 0
            snrs = np.nan_to_num(self._parent.get_snr2(epochs = self.epochs,alpha=self.alpha, delta=self.delta, psi=self.psi, phi0=self.phi0, cosi=self.cosi, Sn=Sn, det=det, tsft=self.tsft, h0=self.h0,antenna=antenna))
            # set snr to 0 if all elemts of sft power equa to mean in each segment (i.e. no data in this segment)
            mean_val_gap = np.all(sft.norm_sft_power == np.ones(len(sft.norm_sft_power[0]))*2,axis=1)
            #also zero if nan in Sn or data
            Sn_nan = np.isnan(Sn)
            # gen index where there is no data
            normnan = np.all(np.isnan(sft.norm_sft_power),axis=1)
            gapinds = np.any([mean_val_gap,Sn_nan,normnan],axis=0)
            # set that index snr to 0
            snrs[gapinds] = 0

            # loop over all epochs
            update_index = []
            for i,ind_floor in zip(list(range(len(sft.pulsar_index))),sft.pulsar_index):
                
                # set the index of the frequency bin which the signal is in
                pul_freq = sft.pulsar_path[i]
                # convert the index to an actual frequency
                freq_floor = (ind_floor/self.tsft)+self.fmin #+ 0.5/self.tsft
                
                # set limits on which how many bins the signal should be injected into +/- 3 bins of injection bin
                limits = (np.arange(7).astype(float)-3)

                # set the gaps in data to mean of chi2
                if gapinds[i]:
                    sft.norm_sft_power[i] = np.ones(len(sft.norm_sft_power[i]))*2
                    snrsq = 0
                else:
                    snrsq = snrs[i]
                    
                # append calcualtion of snr as a function of time
                snr_calc.append(snrsq)
                if snrsq == 0:
                    continue
                
                # if spectrum already defined have offset based of snr of noise line
                if noise_spect is not None:
                    # get area around signal injection
                    limlist = np.array(ind_floor + limits).astype(int)
                    # get area within band bounds
                    limlist = limlist[limlist > 0]
                    limlist = limlist[limlist < len(sft.norm_sft_power[0])]
                    # take the median of the power in this area, subtract the chi2 mean and divide by power (SNR_{line_signal} = ((P_{line_signal} + P_noise - P_noise)/P_noise)
                    # !!!!!!!!!!!!!!!!!!!currently turned off!!!!!!!!!!!!!!!!!!!!!!!!!
                    noise_offset = 0#np.median(sft.norm_sft_power[i,limlist]) - 2
                    if np.isnan(noise_offset) or noise_offset < 0:
                        noise_offset = 0
                else:
                    noise_offset = 0
                    
                # loop over the area around injection
                for lim in limits:
                    #bin center location
                    bin_cent_freq = freq_floor + lim/self.tsft
                    bin_floor_index = int(ind_floor + lim) #np.floor(((j1 + lim/tsft) - fmin)/tsft).astype(int)

                    if bin_floor_index < 0 or bin_floor_index >= len(sft.norm_sft_power[0]):
                        continue
                    
                    # distribute the power across \pm n bins accoring to the fresnel integral and multiply by the snr for that bin add noise offset if this exists
                    update_index.append([int(i),int(bin_floor_index),bin_cent_freq,pul_freq,snrsq,noise_offset])

            # update bins which have signal in them
            if len(update_index) != 0:
                update_index = np.array(update_index)
                # [time_ind, bin_floor_ind, bin_cent_freq, pulsar_freq, snr_bin, noise_offset]
                ncps = self.fresnel_power(update_index[:,2], update_index[:,3], self.tsft,self.f[1])*update_index[:,4] + update_index[:,5]
                update_index = np.c_[update_index, ncps]
                # [time_ind, bin_floor_ind, bin_cent_freq, pulsar_freq, snr_bin, noise_offset, ncp]
                update_index = update_index[update_index[:,6] !=0 ]
                # get all indicies to be updated and calculate the non central chi2 quared with corresponding non centrality parameter

                if rand:
                    sft.norm_sft_power[tuple(update_index[:,:2].astype(int).T)] = st.ncx2.rvs(int(2), nc=update_index[:,6], loc=0,scale=1)
                elif not rand:
                    sft.norm_sft_power[tuple(update_index[:,:2].astype(int).T)] = update_index[:,6]

        # calcualte recovered snr signal, should be equal to simulated snr
        sft.snr_recovered = np.sqrt(np.sum(np.array(snr_calc)))
        if self.h0 == 0:
            sft.depth = np.inf
            sft.median_Sn = np.nanmedian(Sn)
        else:
            # estimate the depth
            sft.depth = np.sqrt(np.nanmedian(Sn))/self.h0
            sft.median_Sn = np.nanmedian(Sn)
        return sft
    
    
    def sum_sfts(self,sum_type="norm_sft_power",gap_val = 2,nsfts=48):
        '''
        takes input sfts with tsft=1800s and returns sum of 48 sfts, i.e summing over a day
        and returns fraction of real data in each sum.
        args
        -------
        gap_val: double
            value used when there are gaps in the data (for calculating fraction)
        returns
        -----------
        data_av: array
            data summed over every day
        fraction: array
            fraction of each day which contained real data
        '''
        for dt in self.det_names:
            data = getattr(self,dt)
            data.sum_sft(sum_type=sum_type,gap_val=gap_val,nsfts=nsfts,remove_original=False)

    def downsamp_frequency(self,data_type="summed_norm_sft_power",stride = 2,remove_original=False):
        '''
        downsample the frequency by taking the mean of "stride" bins starting at base of band
        args
        -------
        stride: double
            number of frequency bins to take the mean of
        '''
        for dt in self.det_names:
            data = getattr(self,dt)
            data.downsamp_frequency(data_type=data_type,stride=stride,remove_original=remove_original)

            

    

